# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT.
#
# DMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

import os
import _pickle as cpickle
import shutil
import copy
import re
import filecmp
import numpy as np
from pathlib import Path
from joblib import Parallel, delayed
from DMT.core import DutType, print_progress_bar, DutView
from DMT.exceptions import NoOpenDeembeddingDut, NoShortDeembeddingDut

try:
    from pylatex import Section, Subsection, SmallText, Tabular, NoEscape, Center
    from DMT.external.pylatex import Tex
except ImportError:
    pass


class __Filter(object):
    """Superclass for all implemented filters. Is used to compare specified properties of devices and deembedding structures.

    Parameters
    ----------
    devProp     : 'str'
        Property of the device that is to be compared to that of the deembedding structure.
    testProp    : 'str'
        Property of the deembedding structure that is to be compared to that of the device.


    Methods
    -------
    filter(dev, testStructure)
        Compare the specified property of a device to the corresponding one of the teststructure.

    """

    def __init__(self, devProp, testProp=None):
        if testProp is None:
            self.testProp = devProp
        else:
            self.testProp = testProp
        self.devProp = devProp

    def filter(self, dev, testStructure):
        """Compares the device and teststructure with regards to the specified property and returns 'True' if they match.

        Parameters
        ----------
        dev             : [DutMeas]
            Object of a DutMeas class. TODO: Should I make that DutView? So it can also be used for sims?
        testStructure   : [DutMeas]
            Object of DutMeas class with the dut_type 'open' or 'short'.
        """
        try:
            propTest = getattr(testStructure, self.testProp)
            propDev = getattr(dev, self.devProp)
        except AttributeError as err:
            raise IOError(
                "DMT -> DutLib -> Filter: the property " + self.devProp + " is not existent."
            ) from err

        if isinstance(propDev, str):
            if propDev == propTest:
                # if propDev in propTest or propTest in propDev:
                return True
            else:
                return False
        else:
            if np.isclose(propDev, propTest, 1e-8, 1e-8):
                return True
            else:
                return False


class LenFilter(__Filter):
    """Subclass of __Filter, that compares the device's and the testStructure's emitter length."""

    def __init__(self):
        super().__init__("length")


class WidthFilter(__Filter):
    """Subclass of __Filter, that compares the device's and the testStructure's emitter width."""

    def __init__(self):
        super().__init__("width")


class NameFilter(__Filter):
    """Subclass of __Filter, that compares the device's and the testStructure's name (contact configuration)."""

    def __init__(self):
        super().__init__("deemb_name")  # , testProp = "deemb_name")


class LenNameFilter:
    """Class that combines the typically used filters for emitter length and device name."""

    def __init__(self):
        self.nameFilter = NameFilter()
        self.lenFilter = LenFilter()

    def filter(self, dev, testStructure):
        return self.nameFilter.filter(dev, testStructure) & self.lenFilter.filter(
            dev, testStructure
        )


class DutLib(object):
    """DutLib is a class managing a library in which measured DUTs in DMT are contained.

    Class is able to match any device type with its corresponding deembedding dummies and perform the deembedding process.

    Parameters
    ----------
    deem_types      : [:class:`~DMT.core.dut_type.DutType`], optional
        A list of dut_types, that are used to filter the incoming list of duts for those devices that need deembedding.
    AC_filter_names    : [tuple], optional
        List of filter tuples that are used to determine the deembedding file corresponding to one measurement file. (e.g.: [('freq', 'hot'),('Spar', 'cold')],...)
    DC_filter_names    : [tuple], optional
        List of filter tuples that are used to determine the deembedding file corresponding to one measurement file. (e.g.: [('freq', 'hot'),('Spar', 'cold')],...)
    is_deembedded_AC   : bool, optional
        If true, all devices in the library have been AC deembedded.
    is_deembedded_DC   : bool, optional
        If true, all devices in the library have been DC deembedded.
    save_dir        : str, optional
        Here the DutLib will try to save itself
    force           : bool, optional
        If True, a already existing library will be deleted.
    n_jobs          : int, optional
        Number of parallel jobs, passed on to joblib.Parallel.

    Attributes
    ----------
    deem_types      : [:class:`~DMT.core.dut_type.DutType`]
        A list of dut_types, that are used to filter the incoming list of duts for those devices that need deembedding.
    AC_filter_names    : [tuple]
        List of filter tuples that are used to determine the deembedding file corresponding to one measurement file. (e.g.: [('freq', 'hot'),('Spar', 'cold')],...)
    DC_filter_names    : [tuple]
        List of filter tuples that are used to determine the deembedding file corresponding to one measurement file. (e.g.: [('freq', 'hot'),('Spar', 'cold')],...)
    is_deembedded_AC   : bool
        If true, all devices in the library have been AC deembedded.
    is_deembedded_DC   : bool
        If true, all devices in the library have been DC deembedded.
    duts            : [:class:`~DMT.core.dut_view.DutView`]
        The devices that shall be managed by this dut
    dut_ref         : :class:`~DMT.core.dut_view.DutView`
        The reference device of this technology
    dut_intrinsic   : :class:`~DMT.core.dut_view.DutView`
        The intrinsic dut of the reference dut (without rbi)
    dut_internal    : :class:`~DMT.core.dut_view.DutView`
        The internal dut of the reference dut (with rbi)
    save_dir        : str
        Here the DutLib will try to save itself
    n_jobs          : int, optional
        Number of parallel jobs, passed on to joblib.Parallel.

    wafer : int or str
        A unique identifier of the wafer that the data in this lib stems from
    date_tapeout  : str
        Tapeout date.
    date_received : str
        Received date.

    Methods
    -------
    find_devices(type)
        Searches for all devices of the specified type and puts them into a separate list.
    sort_duts()
        Assigns the correct Deembedding structures to the DUT.
    deembed_AC()
        Deembeds measured AC data with the corresponding O&S data.
    deembed_DC()
        Deembeds measured AC data with the corresponding O&S data.
    """

    def __init__(
        self,
        deem_types=None,
        AC_filter_names=None,
        DC_filter_names=None,
        is_deembedded_DC=False,
        is_deembedded_AC=False,
        save_dir=None,
        force=False,
        n_jobs=4,
    ):
        if deem_types is None:
            self.deem_types = [DutType.npn]
        else:
            self.deem_types = deem_types

        self.AC_filter_names = AC_filter_names
        self.DC_filter_names = DC_filter_names
        self.is_deembedded_AC = is_deembedded_AC
        self.is_deembedded_DC = is_deembedded_DC

        self.deem_open = (
            DutType.flag_open
        )  # deem_open_bjt # look only for the flag not for the device!
        self.deem_short = DutType.flag_short  # deem_short_bjt

        self.duts = []  # The devices that shall be managed by this dut
        self._dut_ref = None  # The reference device of this technology
        self.dut_ref_dut_dir = None
        self._dut_intrinsic = None  # The intrinsic dut of the reference dut (without rbi)
        self.dut_intrinsic_dut_dir = None  # The intrinsic dut of the reference dut (without rbi)
        self._dut_internal = None  # The internal dut of the reference dut (with rbi)
        self.dut_internal_dut_dir = None  # The intrinsic dut of the reference dut (without rbi)
        self._save_dir = None  # Here the DutLib will try to save itself
        if save_dir is not None:
            if force:
                try:
                    shutil.rmtree(save_dir)
                except FileNotFoundError:
                    pass

            self.save_dir = save_dir  # Here the DutLib will try to save itself

        self.ignore_duts = []  # list of names which are not returned while iteration

        self.n_jobs = n_jobs  # number of parallel jobs while directory import

        # additional information to help assess the data later
        self.wafer = None
        self.date_tapeout = None
        self.date_received = None

    @property
    def dut_ref(self):
        """Returns always just self._dut_ref"""
        if self._dut_ref is None:
            raise IOError("Dut_ref doesn't exist!")

        return self._dut_ref

    @dut_ref.setter
    def dut_ref(self, dut):
        """Ensure that dut_ref is in duts"""
        if not id(dut) in [id(dut_) for dut_ in self.duts]:
            self.duts.append(dut)

        self._dut_ref = dut

    @property
    def dut_internal(self):
        """Returns always just self._dut_internal"""
        if self._dut_internal is None:
            raise IOError("Dut_internal doesn't exist!")

        return self._dut_internal

    @dut_internal.setter
    def dut_internal(self, dut):
        """Ensure that dut_internal is in duts"""
        if not id(dut) in [id(dut_) for dut_ in self.duts]:
            self.duts.append(dut)

        self._dut_internal = dut

    @property
    def dut_intrinsic(self):
        """Returns always just self._dut_intrinsic"""
        if self._dut_intrinsic is None:
            raise IOError("Dut_intrinsic doesn't exist!")

        return self._dut_intrinsic

    @dut_intrinsic.setter
    def dut_intrinsic(self, dut):
        """Ensure that dut_intrinsic is in duts"""
        if not id(dut) in [id(dut_) for dut_ in self.duts]:
            self.duts.append(dut)

        self._dut_intrinsic = dut

    @property
    def save_dir(self):
        """Just return save dir"""
        return self._save_dir

    @save_dir.setter
    def save_dir(self, path):
        """Ensures an empty folder for the DutLib"""
        path = Path(path).resolve()
        if (path / "dut_lib.p").is_file():
            raise FileExistsError(
                "The path you chose is already used by an other library! Either delete the already existing library or load it."
            )

        self._save_dir = path

    def find_devices(self, devtype):
        """Searches the list of duts read in from the specified folder previously for the specified dut_type and adds all identified devices to a new list.

        Parameters
        ----------
        devtype : tuple
            Contains one or several dut_types that are to be identified to be added to a new list.
        """

        sorted_list = []
        # find specified dut_type and append element
        for ty in devtype:
            for dut in self:
                if dut.dut_type.is_subtype(ty):
                    sorted_list.append(dut)

        return sorted_list

    def import_directory(
        self, import_dir, dut_filter, dut_level=1, temperature_converter=None, force=True, **kwargs
    ):
        """Read in all files in import_dir into a DutLib object.

        Read in all devices in the subfolders of a directory.

        Data "dut_level" folders apart with respect to import_dir will be stored into one database and one DutView object.
        Example folder structure to illustrate this principle::

            Measurement
            ├───die1
            |   ├───Device1
            |   └───Device2
            └───die2
                └───Device1

        then for importing with this method, import_dir is set to `Measurement/` and dut_level is set to 2.
        The DutView objects need to be supplied by the callable object (function) dut_filter, which is supplied by the user.
        The dut_filter function is called with the DutView's relative path with respect to import dir and shall return a fully initialized DutView object.

        Parameters
        ----------
        import_dir       :  string or os.Pathlike
            Path to the directory that contains all dies on a wafer.
        dut_filter       :  callable
            User supplied function that is called with the relative paths dut_level levels below import_dir and returns DutView objects.
        dut_level        :  int
            Subfolder level that contains data of ONE specifid DutView. Files that lie in this level will be put into one database.
        temperature_converter :  callable object, optional
            Called to convert the directory name into a key part.
            If key does not contain a temperature, it should return -1.
            If it does contain a temperature, the temperature in Kelvin should be returned.
            Defaults to :meth:`~DMT.core.dut_meas.DutMeas.temp_converter_default`
        force            :  bool
            Default = True. If True, databases and duts that have already been imported are overwritten.
        kwargs   :  bool
            Additional keyword arguments that are passed to the read_data routines, which are called to read the (mdm, csv or elpa) data. E.g. if you have delimiter ',' in your .csv file, pass delimeter=','.

        Returns
        -------
        duts : [:class:`~DMT.core.dut_view.DutView`]
            DutViews loaded from the given directory
        """
        if not isinstance(import_dir, Path):
            import_dir = Path(import_dir).resolve()

        if not callable(dut_filter):
            raise IOError(
                "DMT -> DataManager -> import_directory(): You did not specify a dut_filter function.\n How shall DMT know what kind of Dut is where?"
            )

        print("\n")
        print("DMT will now try to recursively import DutView objects from directory:")
        print(import_dir)
        print("\n")

        # prepare progress bar and find out how many duts we will read
        duts = []
        dut_paths = []
        if dut_level == 0:
            duts.append(dut_filter(str(import_dir)))
            dut_paths.append(import_dir)
        else:
            for child in import_dir.glob("*/" * (dut_level)):  # can be used in windows and linux
                if child.is_dir():  # only directories are allowed
                    child = child.resolve()
                    dut = dut_filter(str(child))
                    if dut is not None:
                        dut_paths.append(child)
                        duts.append(dut)

        if len(duts) == 0:
            raise IOError(
                "DMT -> collect_data: Did not find a single candidate for importing in directory "
                + str(import_dir)
                + " ."
            )

        # walk through all files of each dut and gather the mdm files
        datas = Parallel(n_jobs=self.n_jobs, verbose=10)(
            _read_dut_folder(dut, path, force, temperature_converter, **kwargs)
            for dut, path in zip(duts, dut_paths)
        )

        for dut, data in zip(duts, datas):
            dut._data = data  # pylint: disable=protected-access

        print("DutLib imported " + str(len(duts)) + " DUTs.")

        if not self.is_deembedded_AC and not self.is_deembedded_DC:
            self.add_duts(duts)

        return duts

    def add_duts(self, duts):
        """Add duts to the DutLib duts array.

        Parameters
        ----------
        duts : [DutView] or DutView
            The dut or duts that shall be added.
        """
        if self.is_deembedded_AC or self.is_deembedded_DC:
            raise IOError("DutLib: I am already deembedded and don" "t want to add Duts.")

        try:
            for dut in duts:
                if dut.name in [dut_a.name for dut_a in self.duts]:
                    raise IOError(
                        "DutLib: A DuT of this name already exists. Dut names must be unique!"
                    )
                self.duts.append(dut)
        except TypeError:
            if duts.name in [dut_a.name for dut_a in self.duts]:
                raise IOError(
                    "DutLib: A DuT of this name already exists. Dut names must be unique!"
                )
            self.duts.append(duts)

    def save(self):
        """Save the DutLib to save_dir."""
        assert self.save_dir is not None

        if not os.path.isabs(self.save_dir):
            self._save_dir = os.path.abspath(self.save_dir)

        if not os.path.exists(self.save_dir):
            os.makedirs(self.save_dir)

        ignore_duts = copy.deepcopy(self.ignore_duts)
        self.ignore_duts = []  # save all duts, even the one who were ignored
        for dut in self.duts:
            if dut.database_dir != os.path.join(self.save_dir, "duts"):
                try:  # if enough data available, sort by wafer and dies
                    directory = os.path.join(
                        self.save_dir, "duts", "wafer_" + str(dut.wafer), "die_" + str(dut.die)
                    )
                except:
                    directory = os.path.join(self.save_dir, "duts")
                dut.database_dir = directory

            dut.save()

        if self._dut_ref is not None:
            path_abs = Path(self.dut_ref.dut_dir).resolve()
            if filecmp.cmp(path_abs, self.dut_ref.dut_dir):
                self.dut_ref_dut_dir = path_abs
            else:
                raise IOError(
                    "DMT->DutLib->Save: I can not convert the relative dut path of dut_ref into an absolute path!"
                )

        if self._dut_internal is not None:
            path_abs = Path(self.dut_internal.dut_dir).resolve()
            if filecmp.cmp(path_abs, self.dut_internal.dut_dir):
                self.dut_internal_dut_dir = path_abs
            else:
                raise IOError(
                    "DMT->DutLib->Save: I can not convert the relative dut path of dut_internal into an absolute path!"
                )

        if self._dut_intrinsic is not None:
            path_abs = Path(self.dut_intrinsic.dut_dir).resolve()
            if filecmp.cmp(path_abs, self.dut_intrinsic.dut_dir):
                self.dut_intrinsic_dut_dir = path_abs
            else:
                raise IOError(
                    "DMT->DutLib->Save: I can not convert the relative dut path of dut_intrinsic into an absolute path!"
                )

        self.ignore_duts = ignore_duts  # but save the list ignore
        with open(os.path.join(self.save_dir, "dut_lib.p"), "wb") as handle:
            cpickle.dump(self, handle)

    @staticmethod
    def load(lib_directory):
        """Static class method. Loads a DutLib object from a pickle file with full path lib_directory.

        Parameters
        ----------
        lib_directory  :  str or os.Pathlike
            Path to the direcotry that contains a pickled DutLib object that shall be loaded.

        Returns
        -------
        obj  :  DutLib()
            Loaded object from the pickle file.
        """
        # pylint: disable=unused-variable
        lib_directory = Path(lib_directory).resolve()
        with (lib_directory / "dut_lib.p").open(mode="rb") as handle:
            dut_lib = cpickle.load(handle)
            dut_lib._save_dir = Path(dut_lib.save_dir)  # pylint: disable=protected-access

        save_dir_old = ""
        # need to cast paths? This is needed when the machine is changed -> new absolute paths...
        if (
            lib_directory != dut_lib.save_dir
        ):  # dut_lib.save_dir is the old absolute path of the lib
            save_dir_old = str(dut_lib.save_dir)
            # write to internal value to avoid consistency check
            dut_lib._save_dir = lib_directory  # pylint: disable=protected-access

        # load all duts
        ignore_duts = copy.deepcopy(dut_lib.ignore_duts)
        dut_lib.ignore_duts = []  # load all duts, even the one who were ignored
        for dir_name, dir_list, file_list in os.walk(dut_lib.save_dir / "duts"):  # find the duts
            for file_dut in file_list:
                if file_dut.endswith(".p"):
                    dut_lib.duts.append(DutView.load_dut(os.path.join(dir_name, file_dut)))

        # correct dut paths:
        if dut_lib.dut_ref_dut_dir is not None:
            dut_lib.dut_ref_dut_dir = Path(dut_lib.dut_ref_dut_dir)
            if not dut_lib.dut_ref_dut_dir.exists():
                dut_lib.dut_ref_dut_dir = Path(
                    str(dut_lib.dut_ref_dut_dir).replace(save_dir_old, str(lib_directory))
                )
        if dut_lib.dut_internal_dut_dir is not None:
            dut_lib.dut_internal_dut_dir = Path(dut_lib.dut_internal_dut_dir)
            if not dut_lib.dut_internal_dut_dir.exists():
                dut_lib.dut_internal_dut_dir = Path(
                    str(dut_lib.dut_internal_dut_dir).replace(save_dir_old, str(lib_directory))
                )
        if dut_lib.dut_intrinsic_dut_dir is not None:
            dut_lib.dut_intrinsic_dut_dir = Path(dut_lib.dut_intrinsic_dut_dir)
            if not dut_lib.dut_intrinsic_dut_dir.exists():
                dut_lib.dut_intrinsic_dut_dir = Path(
                    str(dut_lib.dut_intrinsic_dut_dir).replace(save_dir_old, str(lib_directory))
                )

        for dut in dut_lib.duts:
            if dut_lib.dut_ref_dut_dir is not None:
                if filecmp.cmp(dut.dut_dir, dut_lib.dut_ref_dut_dir):
                    dut_lib.dut_ref = dut
            if dut_lib.dut_internal_dut_dir is not None:
                if filecmp.cmp(dut.dut_dir, dut_lib.dut_internal_dut_dir):
                    dut_lib.dut_internal = dut
            if dut_lib.dut_intrinsic_dut_dir is not None:
                if filecmp.cmp(dut.dut_dir, dut_lib.dut_intrinsic_dut_dir):
                    dut_lib.dut_intrinsic = dut

        dut_lib.ignore_duts = ignore_duts  # but keep the list ignore
        return dut_lib

    def __getstate__(self):
        """Return state values to be pickled. Implemented according `to <https://www.ibm.com/developerworks/library/l-pypers/index.html>`_ ."""
        d = copy.copy(self.__dict__)
        if "duts" in d:
            del d["duts"]
        if "dut_ref" in d:
            del d["dut_ref"]
        if "dut_intrinsic" in d:
            del d["dut_intrinsic"]
        if "dut_internal" in d:
            del d["dut_internal"]
        return d

    def __setstate__(self, state):
        """Return state values to be pickled. Implemented according `to <https://www.ibm.com/developerworks/library/l-pypers/index.html>`_ ."""
        # pylint: disable = attribute-defined-outside-init
        self.__dict__ = state
        self.__dict__["duts"] = []
        self.__dict__["dut_ref"] = None
        self.__dict__["dut_intrinsic"] = None
        self.__dict__["dut_internal"] = None

    def deembed_AC(self, width_filter, length_filter, name_filter, user_fun=None):
        """Assign O&S structures to the correct devices for later deembedding process and adds the thus assigned devices to a new list,
        which only includes teststructures for characterization and no O&S structures separately.

        Parameters
        ----------
        width_filter : bool
            If true: use de-embedding structures at same width, else ignore width.
        length_filter : bool
            If true: use de-embedding structures at same length, else ignore length.
        name_filter : bool
            If true: use de-embedding structures with same string in property 'deemb_name', else ignore.
        user_fun : callable, optional
            User specific function that gets one Dut object as an argument and returns the corresponding open, short Duts.

        Methods
        -------
        lowLevelSort(testStru,fn)
            Applies the necessary filter functions to a device and a deembedding structure and adds the correct deembedding structure accordingly to the device.

        Notes
        -----
        ..todo: allow the user to pass his own filters.

        """
        if self.is_deembedded_AC:
            raise IOError("Library has already been deembedded.")

        lenFilter = LenFilter()
        widthFilter = WidthFilter()
        nameFilter = NameFilter()

        # Add duts to separate lists according to dut_type.
        dev_list = self.find_devices(self.deem_types)

        if len(dev_list) == 0:
            print("Warning: No devices in library require AC de-embedding.")
            return
        open_list = self.find_devices([self.deem_open])
        short_list = self.find_devices([self.deem_short])

        if len(short_list) == 0 or len(open_list) == 0:
            raise IOError("DMT -> DutLib -> sort_duts: Did not find any open or short duts.")

        # in the simplest case we only have one deemb device. Then use it.
        if len(open_list) == 1 and len(short_list) == 1:
            print("\n")
            print("DMT will now try to deembed all devices in DutLib object:")
            for i_dev, dev in enumerate(dev_list):
                self.deembed_dut_AC(dev, open_list[0], short_list[0])
                print_progress_bar(
                    i_dev, len(dev_list), prefix="Deembedding AC:", suffix=dev.name, length=50
                )

            print_progress_bar(
                len(dev_list), len(dev_list), prefix="Deembedding AC:", suffix=dev.name, length=50
            )
            self.is_deembedded_AC = True
            return

        # more than one deemb device, then we need filters.
        if not width_filter and not length_filter and not name_filter and user_fun is None:
            raise IOError(
                "DMT -> DutLib: You did not select any filter flags that would allow deembedding! Alternative: pass user_fun"
            )

        # Iterating through the dev_list, which contains all devices that require deembedding.
        for i_dev, dev in enumerate(dev_list):
            print("\n")
            print("DMT will now try to deembed all devices in DutLib object:")
            suitable_opens = []
            suitable_shorts = []
            if user_fun is None:
                for deem_dut in open_list + short_list:
                    if name_filter:
                        if not nameFilter.filter(dev, deem_dut):
                            continue

                    if width_filter:
                        if not widthFilter.filter(dev, deem_dut):
                            continue

                    if length_filter:
                        if not lenFilter.filter(dev, deem_dut):
                            continue

                    if deem_dut.dut_type.is_subtype(self.deem_open):
                        suitable_opens.append(deem_dut)
                    else:
                        suitable_shorts.append(deem_dut)
            else:  # user user supplied function
                suitable_open, suitable_short = user_fun(dev)
                suitable_opens.append(suitable_open)
                suitable_shorts.append(suitable_short)

            # prefer devices at same die if multiple options
            if len(suitable_opens) > 1 or len(suitable_shorts) > 1:
                die_target = dev.die
                for dut_open in suitable_opens:
                    if dut_open.die == die_target:
                        suitable_opens = [dut_open]
                        break
                for dut_short in suitable_shorts:
                    if dut_short.die == die_target:
                        suitable_shorts = [dut_short]
                        break

            # if we found more than one suitable short or open, throw an error.
            if len(suitable_opens) > 1 or len(suitable_shorts) > 1:
                message = "DMT -> DutLib -> sort_duts: For dut:\n"
                message += dev.name + "\n"
                message += " more than one short/open deembeding structure was found:\n"
                for suitable_dut in suitable_opens + suitable_shorts:
                    message += suitable_dut.name + "\n"

                raise IOError(message)

            elif len(suitable_opens) == 0:
                raise NoOpenDeembeddingDut(
                    "For "
                    + dev.name
                    + " no open was found. Deemb_name is "
                    + dev.deemb_name
                    + ".\n"
                    + "Available opens: "
                    + "".join([dut.name + " " for dut in open_list])
                )
            elif len(suitable_shorts) == 0:
                raise NoShortDeembeddingDut("For " + dev.name + " no Short was found.")
            else:
                print_progress_bar(
                    i_dev, len(dev_list), prefix="Deembedding AC:", suffix=dev.name, length=50
                )
                self.deembed_dut_AC(dev, suitable_opens[0], suitable_shorts[0])

            print_progress_bar(
                len(dev_list), len(dev_list), prefix="Deembedding AC:", suffix=dev.name, length=50
            )

            # Deembed DC data if required
        self.is_deembedded_AC = True

    def deembed_DC(
        self,
        width_filter,
        length_filter,
        name_filter,
        function_dut=None,
        function_df=None,
        shorts=None,
        t_ref=300,
        forced_current=False,
    ):
        """Assign O&S structures to the correct devices for later deembedding process and adds the thus assigned devices to a new list,
        which only includes teststructures for characterization and no O&S structures separately.

        Parameters
        ----------
        duts : [duts]
            List of duts that needs to be sorted.
        shorts : [duts], None
            List of shorts that should be used.
        function_dut : function, optional
            a function method(dut_short) that returns the metallization resistances as a dict {'R_EM':float64, 'R_BM':float64, 'R_CM':float64}. Use this if the default DMT DC deembeding is not suitable.
        function_df  : function, optional
            a function method(df_short) that returns the metallization resistances as a dict {'R_EM':float64, 'R_BM':float64, 'R_CM':float64}. Use this if the default DMT DC deembeding is not suitable.
        t_ref       : float, optional
            Temperature of the metallization resistances to return

        Methods
        -------
        lowLevelSort(testStru,fn)
            Applies the necessary filter functions to a device and a deembedding structure and adds the correct deembedding structure accordingly to the device.

        Returns
        -------
        mres : dict
            {'R_EM':float64, 'R_BM':float64, 'R_CM':float64}

        Notes
        -----
        ..todo: allow the user to pass his own filters.

        """
        if self.is_deembedded_DC:
            raise IOError("Library has already been deembedded.")

        lenFilter = LenFilter()
        widthFilter = WidthFilter()
        nameFilter = NameFilter()

        # Add duts to separate lists according to dut_type.
        dev_list = self.find_devices(self.deem_types)
        if len(dev_list) == 0:
            print("Warning: No devices in library require DC de-embedding.")
            return
        if shorts is None:
            short_list = self.find_devices([self.deem_short])
        else:
            short_list = shorts

        if len(short_list) == 0 and not function_dut and not function_df:
            raise IOError("DMT -> DutLib -> sort_duts: Did not find any short duts.")

        # in the simplest case we only have one deemb device. Then use it.
        if len(short_list) == 1 or len(short_list) == 0:
            print("\n")
            print("DMT will now try to deembed all devices in DutLib object:")
            short_i = None
            if len(short_list) == 1:
                short_i = short_list[0]

            for i_dev, dev in enumerate(dev_list):
                if dev == self.dut_ref:
                    mres = self.deembed_dut_DC(
                        dev,
                        short_i,
                        function_dut=function_dut,
                        function_df=function_df,
                        t_ref=t_ref,
                        forced_current=forced_current,
                    )
                else:
                    self.deembed_dut_DC(
                        dev,
                        short_i,
                        function_dut=function_dut,
                        function_df=function_df,
                        t_ref=t_ref,
                        forced_current=forced_current,
                    )
                print_progress_bar(
                    i_dev, len(dev_list), prefix="Deembedding DC:", suffix=dev.name, length=50
                )

            print_progress_bar(
                len(dev_list), len(dev_list), prefix="Deembedding DC:", suffix="finish", length=50
            )
            self.is_deembedded_DC = True
            return mres

        # more than one deemb device, then we need filters.
        if not width_filter and not length_filter and not name_filter:
            raise IOError(
                "DMT -> DutLib: You did not select any filter flags that would allow deembedding!"
            )

        # Iterating through the dev_list, which contains all devices that require deembedding.
        for i_dev, dev in enumerate(dev_list):
            print("\n")
            print("DMT will now try to DC deembed all devices in DutLib object:")
            suitable_shorts = []
            for deem_dut in short_list:
                if name_filter:
                    if not nameFilter.filter(dev, deem_dut):
                        continue

                if width_filter:
                    if not widthFilter.filter(dev, deem_dut):
                        continue

                if length_filter:
                    if not lenFilter.filter(dev, deem_dut):
                        continue

                suitable_shorts.append(deem_dut)

            # if multiple shorts, prefer at same die
            if len(suitable_shorts) > 1:
                die_target = dev.die
                for dut_short in suitable_shorts:
                    if dut_short.die == die_target:
                        suitable_shorts = [dut_short]
                        break

            # if we found more than one suitabel short or open, throw an error.
            if len(suitable_shorts) > 1:
                message = r"DMT -> DutLib -> sort_duts: For dut:\n"
                message += dev.name + r"\n"
                message += r" more than one short/open deembeding structure was found:\n"
                for suitable_dut in suitable_shorts:
                    message += suitable_dut.name + r"\n"

                raise IOError(message)

            elif len(suitable_shorts) == 0:
                raise NoShortDeembeddingDut("For " + dev.name + " no short was found.")
            else:
                print_progress_bar(
                    i_dev, len(dev_list), prefix="Deembedding DC:", suffix=dev.name, length=50
                )
                mres = self.deembed_dut_DC(
                    dev,
                    suitable_shorts[0],
                    function_dut=function_dut,
                    function_df=function_df,
                    t_ref=t_ref,
                    forced_current=forced_current,
                )

        print_progress_bar(
            len(dev_list), len(dev_list), prefix="Deembedding DC:", suffix=dev.name, length=50
        )

        self.is_deembedded_DC = True

        return mres

    def deembed_dut_AC(self, dut, dut_open, dut_short):
        """Checks for all AC measurement data that needs to be deembedded. Differentiates between hot and cold S-parameters and picks the deembedding files accordingly.

        Parameters
        ----------
        dut         : [dut]
            Current DuT which is to be deembedded.
        dut_open    : [dut.DutType.open_deem]
            Corresponding open
        dut_short    : [dut.DutType.short_deem]
            Corresponding short

        """
        # Filter function that creates a temporary filter for each deembedding - DuT df pair, which need to be considered for deembedding
        # def temp_filter(filter_name, key_):
        #     if re.search(filter_name, key_, re.IGNORECASE):
        #         return True
        #     else:
        #         return False

        # Go through all available filters and find dfs matching their values
        for key in dut.data.keys():
            requires_deemb = False
            deembedded = False

            # first find out, if any filter applies here. if yes, likely this data needs deembedding
            for (meas_filter, deem_filter) in self.AC_filter_names:
                if re.search(meas_filter, key, re.IGNORECASE):
                    requires_deemb = True
                    break

            for (meas_filter, deem_filter) in self.AC_filter_names:
                # check if df needs to be AC deembedded & find O&S structure
                if re.search(meas_filter, key, re.IGNORECASE):
                    short_keys = []
                    open_keys = []

                    # find all possible opens
                    for open_key in dut_open.data.keys():
                        if re.search(deem_filter, open_key, re.IGNORECASE):
                            open_keys.append(open_key)

                    for short_key in dut_short.data.keys():
                        if re.search(deem_filter, short_key, re.IGNORECASE):
                            short_keys.append(short_key)

                    if len(short_keys) == 0 or len(open_keys) == 0:
                        continue  # maybe another filter applies

                    # if only one suitable short or open has been found we just take it
                    if len(short_keys) == 1 and len(open_keys) == 1:
                        df_open = dut_open.data[open_keys[0]]
                        df_short = dut_short.data[short_keys[0]]
                        dut.data[key] = dut.data[key].deembed(
                            df_open,
                            df_short,
                            ports=dut.nodes,
                            ndevices=dut.ndevices,
                            ndevices_open=dut_open.ndevices,
                            ndevices_short=dut_short.ndevices,
                        )
                        # we get the number of deembedded dirty...
                        # times = dut.ndevices/dut_open.ndevices # not implemented anymore
                        times = 1
                        dut.open_deembedded_with = str(times) + "x" + dut_open.name
                        dut.short_deembedded_with = str(times) + "x" + dut_short.name

                    # bad news...try to find same temperatures. #TODO: Does not work if no keys are found
                    else:
                        key_temperature = dut.get_key_temperature(key)
                        for open_key in open_keys:
                            if key_temperature == dut_open.get_key_temperature(open_key):
                                break

                        for short_key in short_keys:
                            if key_temperature == dut_short.get_key_temperature(short_key):
                                break

                        df_open = dut_open.data[open_key]
                        df_short = dut_short.data[short_key]
                        try:
                            dut.data[key] = dut.data[key].deembed(
                                df_open,
                                df_short,
                                ports=dut.nodes,
                                ndevices=dut.ndevices,
                                ndevices_open=dut_open.ndevices,
                                ndevices_short=dut_short.ndevices,
                            )
                        except ValueError as err:
                            raise IOError(
                                "The dataframes with keys "
                                + open_key
                                + " and "
                                + short_key
                                + " from open device "
                                + dut_open.name
                                + " and short device "
                                + dut_short.name
                                + " are not matching the data in key "
                                + key
                                + " from dut "
                                + dut.name
                                + "."
                            ) from err

                        # we get the number of deembedded dirty...
                        # times = dut.ndevices/dut_open.ndevices
                        times = 1
                        dut.open_deembedded_with = str(times) + "x" + dut_open.name
                        dut.short_deembedded_with = str(times) + "x" + dut_short.name

                    deembedded = True
                    break

            if not deembedded and requires_deemb:
                raise IOError(
                    "During AC Deembedding: Dataframe "
                    + key
                    + " of DutMeas "
                    + dut.name
                    + " seems to require deembedding, but not a single suitable key was found."
                )

    def load_database(self, database_dir, only_meas=True):
        """Loads all DuTs from a given database directory. Does NOT load the data of the duts, they are loaded using run_and_read

        Parameters
        ----------
        database_dir : str
        only_meas : {True, False}, optional
            If True, only folders without "_hash_" are loaded. This is exclusive for :class:`~DMT.core.dut_meas.DutMeas`

        Returns
        -------
        duts : [:class:`~DMT.core.dut_view.DutView`]
            DutViews loaded from the given directory
        """
        if only_meas:
            views = "all measurement duts"
        else:
            views = "all duts"

        print("\n")
        print("DMT will load " + views + " from the database directory:")
        print(database_dir)
        print("\n")

        duts = []
        for dir_dut in os.listdir(database_dir):
            if only_meas:
                if "_hash_" in dir_dut:
                    continue

            duts.append(DutView.load_dut(os.path.join(database_dir, dir_dut, "dut.p")))

        print("DMT loaded " + str(len(duts)) + " DUTs.")

        if not self.is_deembedded_AC and not self.is_deembedded_DC:
            self.add_duts(duts)

        return duts

    # Makes it possible to iterate over DutLib objects
    def __iter__(self):
        return iter([dut for dut in self.duts if dut.name not in self.ignore_duts])

    def __getitem__(self, key):
        return self.duts[key]

    def normalize(self, dut_type):
        for dut in self.duts:
            if dut.dut_type == dut_type:
                if dut.ndevices > 1:
                    for key in dut.data.keys():
                        df = dut.data[key]
                        dut.data[key] = df.parallel_norm(dut.ndevices, *dut.ac_ports)

    def deembed_dut_DC(
        self, dut, dut_short, function_dut=None, function_df=None, t_ref=300, forced_current=False
    ):
        """Deembeds all DC_measurements in GSG-Pads.

        Parameters
        ----------
        dut         : [dut]
            Current DuT which is to be deembedded.
        dut_short   : [dut.DutType.short_deem]
            Corresponding short
        function_dut : function, optional
            a function method(dut_short) that returns the metallization resistances as a dict {'R_EM':float64, 'R_BM':float64, 'R_CM':float64}. Use this if the default DMT DC deembeding is not suitable.
        function_df  : function, optional
            a function method(df_short) that returns the metallization resistances as a dict {'R_EM':float64, 'R_BM':float64, 'R_CM':float64}. Use this if the default DMT DC deembeding is not suitable.
        t_ref       : float, optional
            Temperature of the metallization resistances to return

        Returns
        -------
        dict  :  {'R_EM':float64, 'R_BM':float64, 'R_CM':float64}
        """
        # test input:
        if function_dut is not None and function_df is not None:
            raise IOError(
                "DMT->DutLib->deembed_DC: Either function_dut or function_df can be given, but not both!"
            )

        # print how exactly deembedding is performed for possible future debugging purposes
        # ISSUE: No for-loop "(meas_filter, deem_filter) in self.DC_filter_names:" performed here.
        # all measurements stored under keys are DC deembedded here!
        if function_dut is not None:
            mres_tref = function_dut(dut_short)
            print("\n")
            print(dut.name)
            print("\n")
            for key in dut.data.keys():
                print(key)
                dut.data[key] = dut.data[key].deembed_DC(
                    mres=mres_tref, forced_current=forced_current
                )
        else:
            mres_tref = None

            for (meas_filter, deem_filter) in self.DC_filter_names:
                # get the short keys
                short_keys = []
                for short_key in dut_short.data.keys():
                    if re.search(deem_filter, short_key, re.IGNORECASE):
                        short_keys.append(short_key)

                if len(short_keys) == 1:
                    # if only one short key has been found we just take it for everything
                    df_short = dut_short.data[short_keys[0]]

                    if function_df is None:
                        mres_tref = df_short.determine_mres(forced_current=forced_current)
                    else:
                        mres_tref = function_df(dut_short)

                for key in dut.data.keys():
                    # check if df needs to be DC deembedded
                    if re.search(meas_filter, key, re.IGNORECASE):

                        if len(short_keys) == 1:
                            mres = mres_tref
                        else:  # NOT TESTED!!!!
                            # bad news...try to find matching temperatures. #TODO: Does not work if no keys are found
                            key_temperature = dut.get_key_temperature(key)
                            for short_key in short_keys:
                                if np.isclose(
                                    key_temperature, dut_short.get_key_temperature(short_key)
                                ):
                                    break
                            df_short = dut_short.data[short_key]

                            if function_df is None:
                                try:
                                    mres = df_short.determine_mres(forced_current=forced_current)
                                except IOError as err:
                                    raise IOError(
                                        "Column missing in df of dut "
                                        + dut_short.name
                                        + " of df with key "
                                        + short_key
                                        + ". Available keys: "
                                        + str(df_short.data.keys())
                                        + "."
                                    ) from err
                            else:
                                mres = function_df(dut_short)

                            if np.isclose(key_temperature, t_ref):
                                mres_tref = mres

                        dut.data[key] = dut.data[key].deembed_DC(
                            mres=mres, forced_current=forced_current
                        )

        return mres_tref

    def toTex(self):
        """This function generates a TeX representation of a DutLib.

        This function generates a section with the title "Measured Devices"
        For each DutType a table is generated that summarizes the available device dimensions for this DutType.
        """
        doc = Tex()
        with doc.create(Section("Measured Devices")):
            with doc.create(Subsection("Geometry Overview")):
                dut_types = list(set([dut.dut_type for dut in self]))
                for dut_type in dut_types:
                    try:
                        flavors = [dut.flavor for dut in self if dut.dut_type == dut_type]
                    except AttributeError:
                        for dut in self:
                            if not hasattr(dut, "flavor"):
                                dut.flavor = None
                        flavors = [dut.flavor for dut in self if dut.dut_type == dut_type]
                    flavors = list(set(flavors))  # cast to unique

                    for flavor in flavors:
                        configs = [
                            dut.contact_config
                            for dut in self
                            if dut.dut_type == dut_type and dut.flavor == flavor
                        ]
                        configs = list(set(configs))  # cast to unique
                        for config in configs:
                            if config is None:
                                str_flavor = (
                                    ""
                                    if flavor is None
                                    else " and flavor " + str(flavor).replace("_", r"\_")
                                )
                                doc.append(
                                    NoEscape(
                                        "Measurements for devices of type "
                                        + str(dut_type)
                                        + str_flavor
                                        + " are available with the following geometries:"
                                    )
                                )
                            else:
                                str_flavor = (
                                    ""
                                    if flavor is None
                                    else ", flavor " + str(flavor).replace("_", r"\_")
                                )
                                doc.append(
                                    NoEscape(
                                        "Measurements for devices with device type "
                                        + str(dut_type)
                                        + str_flavor
                                        + " and contact configuration "
                                        + config.replace("_", r"\_")
                                        + " and  "
                                        + " are available with the following geometries:"
                                    )
                                )

                            doc.append("\r")
                            duts = list(
                                set(
                                    [
                                        dut
                                        for dut in self
                                        if dut.dut_type == dut_type
                                        and dut.contact_config == config
                                        and dut.flavor == flavor
                                    ]
                                )
                            )
                            lE0s = list(set([dut.length for dut in duts]))
                            bE0s = list(set([dut.width for dut in duts]))
                            lE0s.sort()
                            bE0s.sort()
                            header = "|" + " c | " * (
                                len(lE0s) + 1
                            )  # one col for each length and one for bE0 indices
                            with doc.create(Center()) as _centered:
                                with doc.create(Tabular(header)) as table:
                                    table.add_hline()
                                    first_row = [
                                        NoEscape(
                                            r"\backslashbox{$b_{\mathrm{E,drawn}}/\si{\micro\meter}$}{$l_{\mathrm{E,drawn}}/\si{\micro\meter}$}"
                                        )
                                    ]
                                    try:
                                        first_row += ["{:04.2f}".format(lE0 * 1e6) for lE0 in lE0s]
                                    except TypeError:
                                        first_row += [
                                            ",".join(
                                                ["{:04.2f}".format(lE0_a * 1e6) for lE0_a in lE0]
                                            )
                                            for lE0 in lE0s
                                        ]

                                    table.add_row(first_row)
                                    table.add_hline()
                                    for bE0 in bE0s:
                                        try:
                                            row = ["{:04.2f}".format(bE0 * 1e6)]
                                        except TypeError:
                                            row = [
                                                ",".join(
                                                    [
                                                        "{:04.2f}".format(bE0_a * 1e6)
                                                        for bE0_a in bE0
                                                    ]
                                                )
                                            ]
                                        for lE0 in lE0s:
                                            try:
                                                # check if dut with these dimensions exists
                                                dut = next(
                                                    dut
                                                    for dut in duts
                                                    if dut.length == lE0 and dut.width == bE0
                                                )
                                                row.append("x")
                                            except StopIteration:
                                                row.append(" ")

                                        table.add_row(row)
                                        table.add_hline()
                            doc.append("\r")

            with doc.create(
                Subsection("Measurement Data over Temperature and Deembedding Structures")
            ):
                # table that gives overview of all devices:
                # for npns:
                # | name  | Measured@T(K) | lE0_drawn | bE0_drawn | Open Deem. Structure | Short Deem. Structure |
                # other:
                # | name  | Measured@T(K) | l_drawn | b_drawn |
                # begin table
                dut_types = list(set([dut.dut_type for dut in self]))
                for dut_type in dut_types:
                    if not dut_type == DutType.npn:
                        continue

                    configs = [dut.contact_config for dut in self if dut.dut_type == dut_type]
                    configs = list(set(configs))  # cast to unique
                    for config in configs:
                        if config is None:
                            doc.append(
                                NoEscape(
                                    r"The following table gives an overview of all measurements for devices of type "
                                    + str(dut_type)
                                    + r"."
                                )
                            )
                        else:
                            doc.append(
                                NoEscape(
                                    r"The following table gives an overview of all devices with contact configuration "
                                    + config.replace("_", r"\_")
                                    + r" and device type "
                                    + str(dut_type)
                                    + r"."
                                )
                            )

                        doc.append("\r")
                        duts = list(
                            set(
                                [
                                    dut
                                    for dut in self
                                    if dut.dut_type == dut_type and dut.contact_config == config
                                ]
                            )
                        )
                        header = "|" + " c | " * 6  # number of columns
                        with doc.create(Center()) as _centered:
                            with doc.create(SmallText()) as _small:
                                with doc.create(Tabular(header)) as table:
                                    table.add_hline()
                                    first_row = [
                                        r"name ",
                                        NoEscape(r"measured @"),
                                        NoEscape(r"$l_{\mathrm{E0,drawn}}$"),
                                        NoEscape(r"$b_{\mathrm{E0,drawn}}$"),
                                        r"Open Deem. Structure",
                                        r"Short Deem. Structure",
                                    ]
                                    table.add_row(first_row)
                                    second_row = [
                                        r"",
                                        NoEscape(r"$T/\si{\kelvin}$"),
                                        NoEscape(r"$/\si{\micro\meter}$"),
                                        NoEscape(r"$/\si{\micro\meter}$"),
                                        r"",
                                        r"",
                                    ]
                                    table.add_row(second_row)
                                    table.add_hline()
                                    for dut in duts:
                                        # find temps
                                        temps = []
                                        for key in dut.data.keys():
                                            temps.append(dut.get_key_temperature(key))
                                        temps = list(set(temps))  # unique
                                        temps.sort()
                                        lE0_drawn = dut.length * 1e6
                                        bE0_drawn = dut.width * 1e6
                                        if len(temps) < 3:
                                            temps_str = ", ".join(["{}"] * len(temps)).format(
                                                *temps
                                            )
                                            row = [
                                                "{:s}".format(dut.name),
                                                temps_str,
                                                "{:04.2f}".format(lE0_drawn),
                                                "{:04.2f}".format(bE0_drawn),
                                                "{:s}".format(dut.open_deembedded_with),
                                                "{:s}".format(dut.short_deembedded_with),
                                            ]
                                            table.add_row(row)
                                            table.add_hline()
                                        else:  # split temps over two lines
                                            temps_str = ", ".join(["{}"] * len(temps[:3])).format(
                                                *temps[:3]
                                            )
                                            row = [
                                                "{:s}".format(dut.name),
                                                temps_str,
                                                "{:04.2f}".format(lE0_drawn),
                                                "{:04.2f}".format(bE0_drawn),
                                                "{:s}".format(dut.open_deembedded_with),
                                                "{:s}".format(dut.short_deembedded_with),
                                            ]
                                            table.add_row(row)
                                            temps_str = ", ".join(["{}"] * len(temps[3:])).format(
                                                *temps[3:]
                                            )
                                            row = [
                                                "",
                                                temps_str,
                                                "",
                                                "",
                                                "",
                                                "",
                                            ]
                                            table.add_row(row)
                                            table.add_hline()
                            doc.append("\r")
            # end table

            # reference device
            lE0 = r"{:04.2f}".format(self.dut_ref.length * 1e6)
            bE0 = r"{:04.2f}".format(self.dut_ref.width * 1e6)
            doc.append(
                NoEscape(
                    r"The reference device "
                    + self.dut_ref.name.replace("_", r"\_")
                    + r" is of type "
                    + str(self.dut_ref.dut_type)
                    + " in "
                    + self.dut_ref.contact_config.replace("_", r"\_")
                    + r" configuration and has $l_{\mathrm{E,drawn}}$ of $\SI{"
                    + lE0
                    + r"}{\micro\meter}$ and $b_{\mathrm{E,drawn}}$ of $\SI{"
                    + bE0
                    + r"}{\micro\meter}$."
                )
            )
            doc.append(
                NoEscape(
                    r"\enspace All extraction steps that do not deal with special test structures (like tetrodes) or with multiple device geometries, show measured data of the reference device."
                )
            )

        return doc


@delayed
def _read_dut_folder(dut, path, force, temperature_converter, **kwargs):
    """Reads all files inside this path for the given dut. Returns the read dictionary.

    Parameter
    -----------
    dut : DutView
    path : str
    force : bool
    temperature_converter : callable or None
    **kwargs
        Passed to DutView.add_data

    Returns
    -------
    dut.data : {key: DMT.DataFrame}
    """
    if not force:
        dut.load_db()
    for root, _dirs, files in os.walk(path):
        for name in files:
            # get extension name and cast to lower case
            extension = name.split(".")[-1]
            extension = extension.lower()
            if (
                (extension == "mdm")
                or (extension == "elpa")
                or (extension == "csv")
                or (extension == "feather")
            ):
                path_root = Path(root)
                # cut the everything before dut lvl and split the path into groups
                key_list = path_root.relative_to(path).parts
                # join the groups together into a valid dut_data key
                key = dut.join_key_temperature(
                    *key_list, name.split(".")[0], temperature_converter=temperature_converter
                )
                dut.add_data(path_root / name, key, force, **kwargs)

    return dut.data



""" Testing the database manager
"""
import pytest
import numpy as np
import pandas as pd
from pathlib import Path
from DMT.config import DATA_CONFIG
from DMT.core import read_data, DataFrame, DatabaseManager

folder_path = Path(__file__).resolve().parent


def test_data_save_load_df():
    db_manager = DatabaseManager()

    df = read_data(folder_path / "test_data" / "Spar_vb.mdm")

    path_tmp = "test/tmp/db_manager/df.p"
    db_manager.save_df(df, path_tmp)

    df_load = db_manager.load_df(path_tmp)

    assert np.all(np.isclose(df_load["vbe"], df["vbe"]))

    db_manager.del_db(path_tmp)


def test_data_save_load_db_hdf():
    temp = DATA_CONFIG["useHDF5Store"]
    DATA_CONFIG["useHDF5Store"] = True
    db_manager = DatabaseManager()

    df_0 = read_data(folder_path / "test_data" / "Spar_vb.mdm")
    df_1 = read_data(folder_path / "test_data" / "Spar_vb.mdm")
    df_2 = read_data(folder_path / "test_data" / "Spar_vb.mdm")

    # need to clean for db save and load
    nodes = ["B", "C", "E", "S"]
    df_0 = df_0.clean_data(
        nodes,
        "E",
        fallback={
            "S_deemb(1,1)": None,
            "S_deemb(2,1)": None,
            "S_deemb(1,2)": None,
            "S_deemb(2,2)": None,
        },
        ac_ports=["B", "C"],
    )
    df_1 = df_1.clean_data(
        nodes,
        "E",
        fallback={
            "S_deemb(1,1)": None,
            "S_deemb(2,1)": None,
            "S_deemb(1,2)": None,
            "S_deemb(2,2)": None,
        },
        ac_ports=["B", "C"],
    )
    df_2 = df_2.clean_data(
        nodes,
        "E",
        fallback={
            "S_deemb(1,1)": None,
            "S_deemb(2,1)": None,
            "S_deemb(1,2)": None,
            "S_deemb(2,2)": None,
        },
        ac_ports=["B", "C"],
    )

    db = {
        "df_0": df_0,
        "df_1": df_1,
        "df_2": df_2,
    }

    path_tmp = "test/tmp/db_manager/df.hdf"
    db_manager.save_db(path_tmp, db)

    db_load = db_manager.load_db(path_tmp)

    assert np.all(np.isclose(db["df_0"]["V_B"], db_load["df_0"]["V_B"]))

    db_manager.del_db(path_tmp)
    DATA_CONFIG["useHDF5Store"] = temp


def test_data_save_load_db_pickle():
    temp = DATA_CONFIG["useHDF5Store"]
    DATA_CONFIG["useHDF5Store"] = False
    db_manager = DatabaseManager()

    df_0 = read_data(folder_path / "test_data" / "Spar_vb.mdm")
    df_1 = read_data(folder_path / "test_data" / "Spar_vb.mdm")
    df_2 = read_data(folder_path / "test_data" / "Spar_vb.mdm")

    # need to clean for db save and load
    nodes = ["B", "C", "E", "S"]
    df_0 = df_0.clean_data(
        nodes,
        "E",
        fallback={
            "S_deemb(1,1)": None,
            "S_deemb(2,1)": None,
            "S_deemb(1,2)": None,
            "S_deemb(2,2)": None,
        },
        ac_ports=["B", "C"],
    )
    df_1 = df_1.clean_data(
        nodes,
        "E",
        fallback={
            "S_deemb(1,1)": None,
            "S_deemb(2,1)": None,
            "S_deemb(1,2)": None,
            "S_deemb(2,2)": None,
        },
        ac_ports=["B", "C"],
    )
    df_2 = df_2.clean_data(
        nodes,
        "E",
        fallback={
            "S_deemb(1,1)": None,
            "S_deemb(2,1)": None,
            "S_deemb(1,2)": None,
            "S_deemb(2,2)": None,
        },
        ac_ports=["B", "C"],
    )

    db = {
        "df_0": df_0,
        "df_1": df_1,
        "df_2": df_2,
    }

    path_tmp = "test/tmp/db_manager/df.p"
    db_manager.save_db(path_tmp, db)

    db_load = db_manager.load_db(path_tmp)

    assert np.all(np.isclose(db["df_0"]["V_B"], db_load["df_0"]["V_B"]))

    db_manager.del_db(path_tmp)
    DATA_CONFIG["useHDF5Store"] = temp


def test_errors():
    # some errors
    db_manager = DatabaseManager()

    with pytest.raises(FileNotFoundError):
        db_manager.load_db("test/tmp/db_manager/not_there.hdf")


def test_alternatives():
    # no dict of dataframes : just pickle it!
    path_tmp = "test/tmp/db_manager/df.hdf"
    db_manager = DatabaseManager()
    db_manager.save_db(path_tmp, {"a": 5})
    dict_test = db_manager.load_db(path_tmp)
    assert dict_test["a"] == 5


def test_pd_frames():
    path_tmp = "test/tmp/db_manager/df.hdf"
    db_manager = DatabaseManager()
    df_0 = read_data(folder_path / "test_data" / "Spar_vb.mdm")
    df_0.__class__ = pd.DataFrame
    db = {"df_0": df_0}

    db_manager.save_db(path_tmp, db)

    assert db["df_0"].__class__ == DataFrame  # side effect :(

    db_load = db_manager.load_db(path_tmp)

    assert db_load["df_0"].__class__ == DataFrame


if __name__ == "__main__":
    test_data_save_load_df()
    test_data_save_load_db_hdf()
    test_data_save_load_db_pickle()
    test_errors()
    test_alternatives()
    test_pd_frames()



""" test ADS input file generation, copy to server, run simulation and copy back
"""
import types
import logging
import numpy as np
from pathlib import Path

from DMT.core import DutType, Sweep, specifiers, SimCon, Plot, MCard
from DMT.core.circuit import (
    Circuit,
    CircuitElement,
    RESISTANCE,
    CAPACITANCE,
    SHORT,
    VOLTAGE,
    HICUML2_HBT,
)

from DMT.xyce import DutXyce

folder_path = Path(__file__).resolve().parent
logging.basicConfig(
    level=logging.DEBUG,
    format="%(levelname)s - %(message)s",
    filename=folder_path.parent.parent / "logs" / "test_xyce_hicum.log",
    filemode="w",
)


def get_circuit(self, use_build_in=False, topology="common_emitter", **kwargs):
    """

    Parameter
    ------------
    circuit_type : str
        For allowed types, see above
    modelcard : :class:`~DMT.hl2.mc_hicum.McHicum`

    Returns
    -------
    circuit : :class:`~DMT.core.circuit.Circuit`

    """
    if use_build_in:
        module_name = HICUML2_HBT
    else:
        module_name = self.default_module_name  # should be "hicuml2va" because vae

    node_emitter = next(f"n_{node.upper()}" for node in self.nodes_list if "E" in node.upper())
    node_base = next(f"n_{node.upper()}" for node in self.nodes_list if "B" in node.upper())
    node_collector = next(f"n_{node.upper()}" for node in self.nodes_list if "C" in node.upper())
    node_substrate = next(f"n_{node.upper()}" for node in self.nodes_list if "S" in node.upper())
    node_temperature = next(f"n_{node.upper()}" for node in self.nodes_list if "T" in node.upper())

    circuit_elements = []
    # model instance
    circuit_elements.append(
        CircuitElement(
            module_name,
            "Q_H",
            [f"n_{node.upper()}" for node in self.nodes_list],
            # ["n_C", "n_B", "n_E"],
            parameters=self,
        )
    )

    if topology == "common_emitter":

        # BASE NODE CONNECTION #############
        # metal resistance between contact base point and real collector
        try:
            rbm = self.get("_rbm").value
        except KeyError:
            rbm = 1e-3

        circuit_elements.append(
            CircuitElement(
                RESISTANCE, "Rbm", ["n_B_FORCED", node_base], parameters=[("R", str(rbm))]
            )
        )
        # shorts for current measurement
        circuit_elements.append(
            CircuitElement(
                SHORT,
                "I_B",
                ["n_BX", "n_B_FORCED"],
            )
        )
        # capacitance since AC already deembeded Rbm
        circuit_elements.append(
            CircuitElement(
                CAPACITANCE, "Cbm", ["n_B_FORCED", node_base], parameters=[("C", str(1))]
            )
        )

        # COLLECTOR NODE CONNECTION #############
        circuit_elements.append(
            CircuitElement(
                SHORT,
                "I_C",
                ["n_CX", "n_C_FORCED"],
            )
        )
        # metal resistance between contact collector point and real collector
        try:
            rcm = self.get("_rcm").value
        except KeyError:
            rcm = 1e-3

        circuit_elements.append(
            CircuitElement(
                RESISTANCE, "Rcm", ["n_C_FORCED", node_collector], parameters=[("R", str(rcm))]
            )
        )
        # capacitance since AC already deembeded Rcm
        circuit_elements.append(
            CircuitElement(
                CAPACITANCE, "Ccm", ["n_C_FORCED", node_collector], parameters=[("C", str(1))]
            )
        )
        # EMITTER NODE CONNECTION #############
        circuit_elements.append(
            CircuitElement(
                SHORT,
                "I_E",
                ["n_EX", "n_E_FORCED"],
            )
        )
        # metal resistance between contact emiter point and real emiter
        try:
            rem = self.get("_rem").value
        except KeyError:
            rem = 1e-3

        circuit_elements.append(
            CircuitElement(
                RESISTANCE, "Rem", ["n_E_FORCED", node_emitter], parameters=[("R", str(rem))]
            )
        )
        # capacitance since AC already deembeded Rcm
        circuit_elements.append(
            CircuitElement(
                CAPACITANCE, "Cem", ["n_E_FORCED", node_emitter], parameters=[("C", str(1))]
            )
        )
        # add sources and thermal resistance
        circuit_elements.append(
            CircuitElement(VOLTAGE, "V_B", ["n_BX", "0"], parameters=[("Vdc", "V_B"), ("Vac", "1")])
        )
        circuit_elements.append(
            CircuitElement(VOLTAGE, "V_C", ["n_CX", "0"], parameters=[("Vdc", "V_C"), ("Vac", "1")])
        )
        circuit_elements.append(
            CircuitElement(VOLTAGE, "V_E", ["n_EX", "0"], parameters=[("Vdc", "V_E"), ("Vac", "1")])
        )

        # metal resistance between contact emitter potential and substrate contact
        if len(self.nodes_list) > 3:
            circuit_elements.append(
                CircuitElement(
                    SHORT,
                    "I_S",
                    ["n_SX", node_substrate],
                )
            )
            try:
                rsm = self.get("_rsm").value
            except KeyError:
                rsm = 5
            circuit_elements.append(
                CircuitElement(RESISTANCE, "R_S", ["n_SX", "n_EX"], parameters=[("R", str(rsm))])
            )
        if len(self.nodes_list) > 4:
            circuit_elements.append(
                CircuitElement(
                    RESISTANCE, "R_t", [node_temperature, "0"], parameters=[("R", "1e9")]
                )
            )
        circuit_elements += [
            "V_B=0",
            "V_C=0",
            "V_S=0",
            "V_E=0",
            "ac_switch=0",
            "V_B_ac=1-ac_switch",
            "V_C_ac=ac_switch",
            "V_S_ac=0",
            "V_E_ac=0",
        ]
    else:
        raise IOError("The circuit type " + topology + " is unknown!")

    return Circuit(circuit_elements)


def get_dut_build_in():
    modelcard = MCard(
        ["C", "B", "E", "S", "T"],
        default_module_name="",
        default_subckt_name="",
        va_file=folder_path / "hicuml2v2p4p0_xyce.va",
    )
    modelcard.load_model_parameters(
        folder_path.parent / "test_core_no_interfaces" / "test_modelcards" / "npn_full.lib",
    )
    modelcard.update_from_vae(remove_old_parameters=True)
    modelcard.get_circuit = types.MethodType(get_circuit, modelcard)
    return DutXyce(
        None,
        DutType.npn,
        modelcard,
        name="Xyce_BI_",
        nodes="C,B,E,S,T",
        copy_va_files=True,
        reference_node="E",
        get_circuit_arguments={"use_build_in": True},
    )


def get_dut_va():
    modelcard = MCard(
        ["C", "B", "E", "S", "T"],
        default_module_name="",
        default_subckt_name="",
        va_file=folder_path / "hicuml2v2p4p0_xyce.va",
    )
    modelcard.load_model_parameters(
        folder_path.parent / "test_core_no_interfaces" / "test_modelcards" / "npn_full.lib",
    )
    modelcard.update_from_vae(remove_old_parameters=True)
    modelcard.get_circuit = types.MethodType(get_circuit, modelcard)
    return DutXyce(
        None,
        DutType.npn,
        modelcard,
        name="Xyce_VA_",
        nodes="C,B,E,S,T",
        copy_va_files=True,
        reference_node="E",
        get_circuit_arguments={"use_build_in": False},
    )


def get_sweep():
    sweepdef = [
        {"var_name": "FREQ", "sweep_order": 4, "sweep_type": "LOG", "value_def": [8, 9, 2]},
        {
            "var_name": specifiers.VOLTAGE + "B",
            "sweep_order": 3,
            "sweep_type": "LIN",
            "value_def": [0, 1, 11],
        },
        {
            "var_name": specifiers.VOLTAGE + "C",
            "sweep_order": 3,
            "sweep_type": "SYNC",
            "master": specifiers.VOLTAGE + "B",
            "offset": specifiers.VOLTAGE + ["C", "B"],
        },
        {
            "var_name": specifiers.VOLTAGE + ["C", "B"],
            "sweep_order": 2,
            "sweep_type": "LIST",
            # "value_def": [0.3, 0, -0.3],
            "value_def": [0],
        },
        {
            "var_name": specifiers.VOLTAGE + "E",
            "sweep_order": 1,
            "sweep_type": "CON",
            "value_def": [0],
        },
    ]
    return Sweep(
        "gummel",
        sweepdef=sweepdef,
        outputdef=[specifiers.CURRENT + "C", specifiers.CURRENT + "B"],
        othervar={"TEMP": 300},
    )


def test_run_and_read():

    col_vbc = specifiers.VOLTAGE + ["B", "C"]
    col_vbe = specifiers.VOLTAGE + ["B", "E"]
    col_ib = specifiers.CURRENT + "B"
    col_ic = specifiers.CURRENT + "C"

    duts_test = []
    duts_test.append(get_dut_build_in())
    duts_test.append(get_dut_va())
    sweep_test = get_sweep()

    sim_con = SimCon(t_max=80)  # damn VA-Compile time
    sim_con.append_simulation(dut=duts_test, sweep=sweep_test)
    sim_con.run_and_read(force=True)

    for dut in duts_test:
        dut_name = dut.name[5:7]
        df = dut.get_data(sweep=sweep_test)

        df.ensure_specifier_column(col_vbe, ports=dut.nodes)
        df.ensure_specifier_column(col_vbc, ports=dut.nodes)

        for _index, vbc, data in df.iter_unique_col(col_vbc, decimals=3):
            data = data[data[specifiers.FREQUENCY] == 1e8]

            if "BI" in dut_name:
                vbe_test = np.array(
                    [
                        4.05656424e-26,
                        1.00000000e-01,
                        2.00000000e-01,
                        3.00000000e-01,
                        4.00000000e-01,
                        5.00000000e-01,
                        5.99999999e-01,
                        6.99999968e-01,
                        7.99998848e-01,
                        8.99985196e-01,
                        9.99951196e-01,
                    ]
                )
                ic_test = np.array(
                    [
                        4.05656424e-23,
                        1.04081409e-13,
                        3.92389201e-13,
                        9.13378816e-12,
                        4.02247275e-10,
                        1.80135598e-08,
                        7.84515642e-07,
                        3.20487174e-05,
                        1.14591374e-03,
                        1.47021196e-02,
                        4.81396797e-02,
                    ]
                )
                ib_test = np.array(
                    [
                        -4.05656424e-23,
                        3.25146606e-13,
                        2.20917953e-12,
                        1.66059841e-11,
                        1.32908623e-10,
                        1.10771760e-09,
                        1.02658644e-08,
                        1.34358642e-07,
                        3.03175108e-06,
                        5.07913940e-05,
                        3.31963182e-04,
                    ]
                )
            elif "VA" in dut_name:
                vbe_test = np.array(
                    [
                        4.05656424e-26,
                        1.00000000e-01,
                        2.00000000e-01,
                        3.00000000e-01,
                        4.00000000e-01,
                        5.00000000e-01,
                        5.99999999e-01,
                        6.99999968e-01,
                        7.99998848e-01,
                        8.99985196e-01,
                        9.99951196e-01,
                    ]
                )
                ic_test = np.array(
                    [
                        4.05656424e-23,
                        1.04081409e-13,
                        3.92389201e-13,
                        9.13378816e-12,
                        4.02247275e-10,
                        1.80135598e-08,
                        7.84515642e-07,
                        3.20487174e-05,
                        1.14591374e-03,
                        1.47021196e-02,
                        4.81396797e-02,
                    ]
                )
                ib_test = np.array(
                    [
                        -4.05656424e-23,
                        3.25146606e-13,
                        2.20917953e-12,
                        1.66059841e-11,
                        1.32908623e-10,
                        1.10771760e-09,
                        1.02658644e-08,
                        1.34358642e-07,
                        3.03175108e-06,
                        5.07913940e-05,
                        3.31963182e-04,
                    ]
                )
            else:
                raise NotImplementedError

            assert np.isclose(np.real(data[col_vbe]), vbe_test).all()
            assert np.isclose(np.real(data[col_ic]), ic_test).all()
            assert np.isclose(np.real(data[col_ib]), ib_test).all()

    return duts_test, sweep_test


if __name__ == "__main__":
    duts, sweep = test_run_and_read()

    col_vbc = specifiers.VOLTAGE + ["B", "C"]
    col_vbe = specifiers.VOLTAGE + ["B", "E"]
    col_ib = specifiers.CURRENT + "B"
    col_ic = specifiers.CURRENT + "C"

    plt_gummel = Plot(
        r"$I_{\mathrm{C}}(V_{\mathrm{BE}})$",
        x_label=r"$V_{\mathrm{BE}}$",
        y_label=r"$I_{\mathrm{C}}$",
        y_log=True,
        style="bw",
    )
    plt_ib = Plot(
        r"$I_{\mathrm{B}}(V_{\mathrm{BE}})$",
        x_label=r"$V_{\mathrm{BE}}$",
        y_label=r"$I_{\mathrm{B}}$",
        y_log=True,
        style="bw",
    )

    for dut in duts:
        dut_name = dut.name[5:7]
        df = dut.get_data(sweep=sweep)

        df.ensure_specifier_column(col_vbe, ports=dut.nodes)
        df.ensure_specifier_column(col_vbc, ports=dut.nodes)

        for _index, vbc, data in df.iter_unique_col(col_vbc, decimals=3):
            data = data[np.isclose(data[specifiers.FREQUENCY], 1e8)]

            plt_gummel.add_data_set(
                np.real(data[col_vbe].to_numpy()),
                np.real(data[col_ic].to_numpy()),
                label=dut_name + " $V_{{BC}} = {0:1.2f} V$".format(data[col_vbc].to_numpy()[0]),
            )
            plt_ib.add_data_set(
                np.real(data[col_vbe].to_numpy()),
                np.real(data[col_ib].to_numpy()),
                label=dut_name + " $V_{{BC}} = {0:1.2f} V$".format(data[col_vbc].to_numpy()[0]),
            )

    plt_ib.plot_pyqtgraph(show=False)
    plt_gummel.plot_pyqtgraph(show=True)



""" Testing the core module mc_parameter
"""
import os
import shutil
import copy
import tempfile
import pytest
import numpy as np
from pathlib import Path
from DMT.core import unit_registry, McParameter, McParameterCollection
from DMT.exceptions import (
    BoundsError,
    ValueTooLargeError,
    ValueExcludedError,
    ParaExistsError,
)
from DMT.external import Tex

folder_path = Path(__file__).resolve().parent
test_path = folder_path.parent


def test_mc_parameter():
    """Testing single parameters"""
    with pytest.raises(IOError):
        param_1 = McParameter(5)

    param_without_bounds = McParameter("a", value=1, minval=None, maxval=None)
    param_with_bounds = McParameter("a", value=1, minval=0, maxval=10)
    param_excl_bounds = McParameter("a", value=1, minval=0, maxval=10)
    param_excl_bounds.inc_max = False
    param_excl_bounds.inc_min = False

    # value type
    param_with_bounds.val_type = int
    with pytest.raises(IOError):
        param_with_bounds.val_type = str
    param_with_bounds.val_type = float

    # setting the minimum
    with pytest.raises(BoundsError):
        param_with_bounds.min = 12
    with pytest.raises(BoundsError):
        param_with_bounds.min = 2

    param_with_bounds.min = 1
    param_with_bounds.min = 0
    with pytest.raises(BoundsError):
        param_excl_bounds.min = 1
    with pytest.raises(BoundsError):
        param_excl_bounds.min = 2

    # setting the maximum
    with pytest.raises(BoundsError):
        param_with_bounds.max = -1
    with pytest.raises(BoundsError):
        param_with_bounds.max = 0.5

    param_with_bounds.max = 1
    param_with_bounds.max = 10
    with pytest.raises(BoundsError):
        param_excl_bounds.max = 1
    with pytest.raises(BoundsError):
        param_excl_bounds.max = 0.5

    # value setting using another mc_parameter
    param_with_bounds.value = param_excl_bounds

    # force set
    param_with_bounds = McParameter("a", value=1, minval=0, maxval=10)
    param_with_bounds._set_forced(1.5)
    assert np.isclose(param_with_bounds.value, 1.5)
    assert np.isclose(param_with_bounds.min, 0)
    assert np.isclose(param_with_bounds.max, 10)

    # forcing a not allow value changes bounds to inf
    param_with_bounds._set_forced(-1.5)
    assert np.isclose(param_with_bounds.value, -1.5)
    assert np.isclose(param_with_bounds.min, -np.inf)
    assert np.isclose(param_with_bounds.max, np.inf)

    param_4 = McParameter("a", value=1, minval=0, maxval=10)

    with pytest.raises(TypeError):
        param_4._set_forced(np.array([-1.5]))

    with pytest.raises(IOError):
        param_4 = McParameter("a", value=1, minval=0, maxval=10)
        param_4._value = -1  # hack
        param_with_bounds._set_forced(param_4)  # forbid to pass on

    # start over and check_bounds
    param_with_bounds = McParameter("a", value=1, minval=0, maxval=10)
    assert param_with_bounds != 1
    assert param_with_bounds != 5

    param_with_bounds.check_bounds(2)
    with pytest.raises(TypeError):
        param_with_bounds.check_bounds(np.array([2]))

    with pytest.raises(TypeError):
        param_with_bounds.check_bounds("test")

    with pytest.raises(ValueTooLargeError):
        param_excl_bounds.check_bounds(10)
    with pytest.raises(ValueTooLargeError):
        param_excl_bounds.check_bounds(11)

    param_excl_bounds.exclude = [3]
    with pytest.raises(ValueExcludedError):
        param_excl_bounds.check_bounds(3)

    # formatting and printing:
    param_format = McParameter("a", value=1, minval=0, maxval=10)
    with pytest.raises(IOError):
        "{:h}".format(param_format)

    assert "{:.1f}".format(param_format) == "1.0"
    assert "{:s}".format(param_format) == "a"
    assert "{:u}".format(param_format) == "-"

    ur = unit_registry
    param_format.unit = ur.ampere
    assert "{:u}".format(param_format) == "\\ampere"
    param_format.unit = ur.dimensionless
    assert "{:u}".format(param_format) == "-"  # same as unit == None

    # None should not be allowed for a float!
    # param_format._value = None # if a parameter is initialized without a value..
    # assert  '{:d}'.format(param_format) == '-'

    # testing iteration: -> should not work
    param_1 = McParameter("a", value=1, minval=0, maxval=10)

    with pytest.raises(TypeError):
        for para in param_1:
            para.value = 1

    # adding two parameter should give an error -> add names or values or append to a composition -> we do not know!
    param_1 = McParameter("a", value=1, minval=0, maxval=10)
    param_2 = McParameter("b", value=1, minval=0, maxval=10)

    with pytest.raises(TypeError):
        _comp = param_1 + param_2


def test_mc_parameter_collection():
    """here compositions are tested!"""
    # create:
    mc_comp = McParameterCollection()

    # add
    mc_comp.add(McParameter("a", value=1, minval=0, maxval=10))

    with pytest.raises(TypeError):
        mc_comp.add("a")

    # add another with same name
    with pytest.raises(ParaExistsError):
        mc_comp.add(McParameter("a", value=2, minval=0, maxval=10))

    with pytest.raises(TypeError):
        _test = mc_comp + 5
    with pytest.raises(TypeError):
        _test = "a" + mc_comp + 5

    # add parameter using magic
    mc_comp = mc_comp + McParameter("b", value=2, minval=0, maxval=10)
    # reversed also possible
    mc_comp = McParameter("c", value=2, minval=0, maxval=10) + mc_comp
    # inplace
    mc_comp += McParameter("d", value=2, minval=0, maxval=10)

    assert len(mc_comp) == 4
    assert mc_comp.name != ["a", "b", "c", "d"]  # all were added in the given order
    assert mc_comp.name == ["c", "a", "b", "d"]  # beware the 'c'

    # to iter them in alphabetical order (needed for hashable netlists):
    for para, name in zip(mc_comp.iter_alphabetical(), ["a", "b", "c", "d"]):
        assert para.name == name

    # adding 2 compositions:
    mc_comp_2 = McParameterCollection()
    mc_comp_2.add(McParameter("x", value=1, minval=0, maxval=10))
    mc_comp_2.add(McParameter("y", value=1, minval=0, maxval=10))
    mc_comp_2.add(McParameter("z", value=1, minval=0, maxval=10))

    assert len(mc_comp + mc_comp_2) == 7
    assert len(mc_comp_2 + mc_comp) == 7
    # order: !
    assert (mc_comp + mc_comp_2).name == ["c", "a", "b", "d", "x", "y", "z"]
    assert (mc_comp_2 + mc_comp).name == ["x", "y", "z", "c", "a", "b", "d"]

    # without magic
    mc_comp.add(mc_comp_2)
    assert len(mc_comp) == 7
    assert mc_comp.name == ["c", "a", "b", "d", "x", "y", "z"]

    mc_comp.remove(mc_comp_2)
    mc_comp.add(mc_comp_2, index=0)
    assert mc_comp.name == ["x", "y", "z", "c", "a", "b", "d"]

    # testing equal:
    # this could also raise an TypeError, but keeping the door open for more magic
    assert mc_comp != "x"
    assert mc_comp != McParameter("x", value=1, minval=0, maxval=10)

    assert mc_comp != mc_comp_2
    assert mc_comp_2 != mc_comp
    mc_comp_copy = copy.deepcopy(mc_comp)
    assert mc_comp == mc_comp_copy

    # setting a new value
    mc_comp.set(McParameter("a", value=3))
    assert mc_comp != mc_comp_copy  # compositions are equal if all parameters and names are equal

    mc_comp.set(mc_comp_copy)  # setting a complete composition is also usefull

    mc_comp["b"] = 10

    with pytest.raises(ValueTooLargeError):
        mc_comp["b"] = 11

    assert mc_comp.get("b").value == 10

    with pytest.raises(KeyError):
        mc_comp.set(McParameter("k", value=3))

    # using the convinience function:
    mc_comp.set_values({"b": 9})
    assert mc_comp.get("b").value == 9

    with pytest.raises(ValueTooLargeError):
        mc_comp.set_values({"b": 11})

    mc_comp.set_values({"b": 11}, force=True)  # can be forced easily :)

    with pytest.raises(KeyError):
        mc_comp.set_values({"g": 11})

    with pytest.raises(KeyError):
        mc_comp.set_values({"g": 11}, force=True)  # force creation of a parameter is removed...

    assert np.isclose(mc_comp["b"].value, 11)
    assert np.isclose(mc_comp["b"].min, -np.inf)
    assert np.isclose(mc_comp["b"].max, np.inf)

    # getting a parameter
    assert mc_comp.get("a").value == 1
    assert mc_comp["a"].value == 1

    assert mc_comp.get_values({"a"})["a"] == 1
    with pytest.raises(KeyError):
        mc_comp.get_values({"a1"})

    # setting new bounds
    with pytest.raises(KeyError):
        mc_comp.set_bounds({"k": (0, 1)})

    mc_comp.set_bounds({"b": (0, 12)})
    assert np.isclose(mc_comp["b"].min, 0)
    assert np.isclose(mc_comp["b"].max, 12)


def test_mc_parameter_collection_to_kwargs():
    """separate test for to_kwargs as this method is very important!!"""
    mc_comp_2 = McParameterCollection()
    mc_comp_2.add(McParameter("x", value=1, minval=0, maxval=10))
    mc_comp_2.add(McParameter("y", value=1, minval=0, maxval=10))
    mc_comp_2.add(McParameter("z", value=1, minval=0, maxval=10))

    control_dict = {"x": 1, "y": 1, "z": 1}
    assert mc_comp_2.to_kwargs() == control_dict


def test_mc_parameter_collection_file_methods():
    """File in and output"""
    mc_comp_2 = McParameterCollection()
    mc_comp_2.add(McParameter("x", value=1, minval=0, maxval=10))
    mc_comp_2.add(McParameter("y", value=1, minval=0, maxval=10))
    mc_comp_2.add(McParameter("z", value=1, minval=0, maxval=10))

    shutil.rmtree(str(test_path / "tmp" / "test_comp"), ignore_errors=True)
    # save and load

    # print part of composition
    assert (
        mc_comp_2.print_parameters(paras=("x", "y"))
        == "  x            = 1.00000e+00   y            = 1.00000e+00 "
    )

    # print to file:
    mc_comp_2.print_to_file(test_path / "tmp" / "test_comp" / "mc_comp_2", create_dir=True)

    assert (
        test_path / "tmp" / "test_comp" / "mc_comp_2.txt"
    ).read_text() == "  x            = 1.00000e+00   y            = 1.00000e+00   z            = 1.00000e+00 \n"


def test_mc_parameter_collection_properties():
    mc_comp_2 = McParameterCollection()
    mc_comp_2.add(McParameter("x", value=1, minval=0, maxval=10, group="g1"))
    mc_comp_2.add(McParameter("y", value=1, minval=0, maxval=10, group="g1"))
    mc_comp_2.add(McParameter("z", value=1, minval=0, maxval=10, group="g2"))

    assert mc_comp_2.group == {"g1", "g2"}  # group is a set of group names

    # set values
    # length must be correct
    with pytest.raises(IOError):
        mc_comp_2.value = [2, 3]
    with pytest.raises(IOError):
        mc_comp_2.value = [2, 3, 4, 5]

    mc_comp_2.value = [2, 3, 4]
    assert all(mc_comp_2.value == np.array([2, 3, 4]))

    # same for min and max
    with pytest.raises(IOError):
        mc_comp_2.min = [2, 3]
    with pytest.raises(IOError):
        mc_comp_2.min = [2, 3, 4, 5]

    mc_comp_2.min = [0, 1, 2]
    assert all(mc_comp_2.min == np.array([0, 1, 2]))

    with pytest.raises(IOError):
        mc_comp_2.max = [2, 3]
    with pytest.raises(IOError):
        mc_comp_2.max = [2, 3, 4, 5]

    mc_comp_2.max = [7, 8, 9]
    assert all(mc_comp_2.max == np.array([7, 8, 9]))

    # remove function:
    mc_comp_2.remove("x")
    assert "x" not in mc_comp_2.name
    mc_comp_2.add(McParameter("x", value=1, minval=0, maxval=10, group="g1"))
    assert "x" in mc_comp_2.name
    mc_comp_2.remove(McParameter("x", value=1))
    assert "x" not in mc_comp_2.name

    mc_comp_2.add(McParameter("x", value=1, minval=0, maxval=10, group="g1"))

    with pytest.raises(KeyError):
        mc_comp_2.remove("a")

    mc_comp_2.remove(("x", "y"))
    mc_comp_2.add(McParameter("x", value=1, minval=0, maxval=10, group="g1"))
    mc_comp_2.add(McParameter("y", value=1, minval=0, maxval=10, group="g1"))

    mc_comp_2.remove(mc_comp_2)

    assert len(mc_comp_2) == 0


def test_mc_parameter_collection_to_tex():
    mc_comp_2 = McParameterCollection()
    mc_comp_2.add(McParameter("x", value=1, minval=0, maxval=10, group="g1"))
    mc_comp_2.add(McParameter("y", value=1, minval=0, maxval=10, group="g1"))
    mc_comp_2.add(McParameter("z", value=1, minval=0, maxval=10, group="g2"))
    mc_comp_2.sort_paras()

    tex = mc_comp_2.print_tex()

    assert isinstance(tex, Tex)

    fp = tempfile.NamedTemporaryFile(mode="w", suffix=".tex", delete=False)  # generate a temp file
    file_name = fp.name
    fp.close()  # it is not deleted

    tex.generate_tex(file_name.replace(".tex", ""))
    with open(file_name, "r") as fp:
        fp.seek(0)
        tex_content = fp.read()

    os.unlink(file_name)  # delete manually -> this needs to be done bc of windows...

    assert "x" in tex_content
    assert "y" in tex_content
    assert "z" in tex_content


def test_json():
    mc_1 = McParameter(
        "x",
        value=1,
        minval=0,
        maxval=None,
        inc_min=True,
        inc_max=False,
        exclude=2,
        group="g1",
        unit=unit_registry.ampere,
        description="asdf",
    )

    target_dict = {
        "name": "x",
        "value": 1.0,
        "min": 0.0,
        "max": np.inf,
        "inc_min": True,
        "inc_max": False,
        "exclude": [2],
        "type": "float",
        "unit": str(unit_registry.ampere),
        "group": "g1",
        "description": "asdf",
        "__McParameter__": "1.0.0",
    }
    assert mc_1.dict_json() == target_dict

    mc_loaded = McParameter.load_json(**target_dict)
    for param in dir(mc_1):
        if not callable(getattr(mc_1, param)) and not param.startswith("__"):
            assert getattr(mc_1, param) == getattr(mc_loaded, param)

    mc_comp = McParameterCollection()
    mc_comp.add(McParameter("x", value=1, minval=0, maxval=10, group="g1"))
    mc_comp.add(McParameter("y", value=1, minval=0, maxval=10, group="g1"))
    mc_comp.add(McParameter("z", value=1, minval=0, maxval=10, group="g2"))

    file_name = "test_json_composition.json"
    mc_comp.dump_json(file_name)
    mc_comp_read = McParameterCollection.load_json(file_name)

    os.remove(file_name)

    assert mc_comp == mc_comp_read


if __name__ == "__main__":
    test_mc_parameter()
    test_mc_parameter_collection()
    test_mc_parameter_collection_to_kwargs()
    test_mc_parameter_collection_file_methods()
    test_mc_parameter_collection_properties()
    test_mc_parameter_collection_to_tex()
    test_json()



""" Testing the hashing class
"""
import os
from DMT.core import create_md5_hash


def test_base():
    file_name = "test.txt"
    with open(file_name, "w") as my_test_file:
        my_test_file.write("test input")

    test_hash = create_md5_hash(file_name, b"example further info")
    assert test_hash == "fd862d8ec7e0ee8cf80aee9408908080"

    test_hash = create_md5_hash(file_name, "example further info")
    assert test_hash == "8242dfc8844947a6aaa3b053e987cf81"

    test_hash = create_md5_hash(file_name, "example further info", 465443)
    assert test_hash == "d3c6394dca24b36e155fc529f187fd17"

    os.remove(file_name)


def test_va_file():
    file_name = "test.txt"
    with open(file_name, "w") as my_test_file:
        my_test_file.write("test input")
    file_name2 = "va_file1.va"
    with open(file_name2, "w") as my_test_file:
        my_test_file.write("test va input 1")
    file_name3 = "va_file2.va"
    with open(file_name3, "w") as my_test_file:
        my_test_file.write("test va input 1")

    test_hash = create_md5_hash(file_name, file_name2)
    assert test_hash == "57649e3096bfd7fd4a17731936e709f7"

    test_hash = create_md5_hash(file_name, file_name2, file_name3)
    assert test_hash == "026beecca8a7ced367e855a484180206"

    os.remove(file_name)
    os.remove(file_name2)
    os.remove(file_name3)


def test_va_content():
    file_name = "test.txt"
    with open(file_name, "w") as my_test_file:
        my_test_file.write("test input")
    file_content1 = "test va input 1"
    file_content2 = "test va input 2"

    test_hash = create_md5_hash(file_name, file_content1)
    assert test_hash == "57649e3096bfd7fd4a17731936e709f7"

    test_hash = create_md5_hash(file_name, file_content1, file_content2)
    assert test_hash == "694a49fc36908036a9aa39b5584e1018"

    os.remove(file_name)


# test case
if __name__ == "__main__":
    test_base()
    test_va_file()
    test_va_content()



""" VA Code structure. As Includes traverse like a tree, it is implemented as a dictionary Oo.
"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

from __future__ import annotations
import zlib, base64
from typing import Union, Callable, Optional
from pathlib import Path
import verilogae

from DMT.core import create_md5_hash


class VACode(object):  # Storage
    """Base file for VA-Codes

    Parameters
    ----------
    name : str
        Name of the code file (should include a relative path to the file which includes this one)
    code : str, optional
        Code of the file
    code_compressed : (str, int), optional
        Compressed code as loaded from a json modelcard file.
    ignore_checksum: bool, optional
        If True, the checksum is ignored, defaults to False.
    """

    def __init__(
        self,
        code: str = "",
        code_compressed: tuple[str, int] = ("", -1),
        ignore_checksum: bool = False,
    ):
        self.code = code

        if not self.code and code_compressed[0]:
            self.decompress_code(
                code_compressed[0], code_compressed[1], ignore_checksum=ignore_checksum
            )

    def __str__(self) -> str:
        """Some short cuts"""
        return self.code

    def __eq__(self, other: VACode) -> bool:
        """Compare two Files. Equal if both name and codes are equal

        Parameters
        ----------
        other : VAFileBase
            Other VAFile to compare

        Returns
        -------
        bool
            True if both have the same name and code...
        """
        if isinstance(other, VACode):
            # the files are equal if the codes are the same ??
            # because if the file are collected from 2 different locations they are still the same -> not true because of includes...
            return self.code == other.code
        elif isinstance(other, str):
            return self.code == other

        return NotImplemented

    @property
    def code_compressed(self):
        """zlib compressed code for saving in modelcard json files.

        Returns
        -------
        (str, str)
            Compressed code and CRC32 check sum after compression and before encoding.
        """
        code = zlib.compress(self.code.encode("utf-8"), level=9)
        csum = zlib.crc32(code)
        code = base64.b85encode(code).decode("utf-8")
        return (code, csum)

    def decompress_code(self, code_compressed: str, csum: int, ignore_checksum: bool = False):
        """Set compressed code to file, will be decompressed and saved in self.code

        Parameters
        ----------
        code_compressed : str
            compressed code
        csum : str
            CRC32 check sum loaded from file.
        ignore_checksum: bool, optional
            If True, the checksum is ignored, defaults to False.

        Raises
        ------
        OSError
            If the saved checksum and checksum of decompressed code do not match
        """
        code = base64.b85decode(code_compressed.encode("utf-8"))
        if not ignore_checksum and csum != zlib.crc32(code):
            raise OSError(
                "Saved checksum and checksum of decompressed code do not match",
                "MCard load compressed va_codes failed. VA-Codes not loaded, manual reset needed!",
            )  # Raise an error here??
        self.code = zlib.decompress(code).decode("utf-8")


class VAFile(object):
    """Tree VA-File for VA-Code. The tree is chosen to correctly mirror possible file structures of multi-file VA-Codes

    Parameters
    ----------
    name : Union[str, Path]
        Absolute or relative path to file or just name of the root file.
    files : dict, optional
        Dictionary with {file_name: VACode}. One of the keys must be the same as the name, by default None
    code : Union[str, VACode], optional
        Code of the root file, by default ""
    code_compressed : (str, int), optional
        Compressed code as loaded from a json modelcard file.
    ignore_checksum: bool, optional
        If True, the checksum is ignored, defaults to False.
    """

    def __init__(
        self,
        name: Union[str, Path],
        files: Optional[dict[str, VACode]] = None,
        code: Union[str, VACode] = "",
        code_compressed: tuple[str, int] = ("", -1),
        ignore_checksum: bool = False,
    ):
        self.files: dict[str, VACode] = {}

        if not isinstance(name, Path):
            name = Path(name)

        self.root: str = name.name
        if files is None:

            if name.is_file():
                self.read_structure(name.parent)
            elif code:
                if isinstance(code, VACode):
                    self.files[self.root] = code
                else:
                    self.files[self.root] = VACode(code=code)
            elif code_compressed[0]:
                self.files[self.root] = VACode(
                    code_compressed=code_compressed, ignore_checksum=ignore_checksum
                )
        else:
            self.files = files

            if self.root not in self.files:
                raise IOError("The root must be part of the file structure.")

    @property
    def root_vfs(self) -> str:
        """Returns the root path inside the vfs

        Returns
        -------
        str
            [description]
        """
        return "/" + self.root

    @property
    def vfs(self) -> dict[str, str]:
        vfs = dict()
        for name, code in self.files.items():
            vfs["/" + name] = str(code)
        return vfs

    def rename_root(self, name_new: str):
        """Rename the root file

        Parameters
        ----------
        name_new : str
            New name

        Raises
        ------
        IOError
            If the new name already existed.
        """
        if name_new in self.files:
            raise IOError("The given new root name already exists in the file tree.")

        self.files[name_new] = self.files[self.root]
        del self.files[self.root]
        self.root = name_new

    def read_structure(self, path_to_own_folder: Union[str, Path]):
        """Checks the file for imports and adds them into the dict. This is done using verilogae.

        Parameters
        ----------
        path_to_own_folder : Union[str, Path]
            Path, in which the main file of the model is located.

        Raises
        ------
        NotImplementedError
            Currently only relative imports are supported. If a file includes an absolute path, the error is raised.
        """
        if not isinstance(path_to_own_folder, Path):
            path_to_own_folder = Path(path_to_own_folder)

        path_to_main_code = path_to_own_folder / self.root

        try:
            self.files[self.root] = VACode(code=path_to_main_code.read_text())  # be safe
        except FileNotFoundError as e:
            va_files = path_to_own_folder.glob("*.va")
            va_files = [str(va_file) for va_file in va_files]
            va_files = "\n" + "\n".join(va_files)
            raise FileNotFoundError(
                e.strerror, e.filename, " . Verilog-A files in this folder:" + va_files
            )

        for file, code in verilogae.export_vfs(str(path_to_main_code)).items():  # type: ignore
            self.files[file[1:]] = VACode(code=code)

    def get_tree_hash(self) -> str:
        """Create a hash for all the codes from this node and all its children. Should be unique for each model...

        Returns
        -------
        str
            MD5 Hash for the code of this file and all its children
        """
        return create_md5_hash(*[vafile.code for vafile in self.files.values()])

    def export_dict(self, compressed_code: bool = False) -> dict:
        """Export self and children into a dictionary for serialization

        Parameters
        ----------
        compressed_code : bool, optional
            Set to true to get compressed code, by default False

        Returns
        -------
        dict
            [description]
        """
        export = {"__root__": self.root}  # this should be not used somewhere!
        for name, vafile in self.iter_codes():
            if compressed_code:
                export[name] = vafile.code_compressed  # type: ignore
            else:
                export[name] = vafile.code

        return export

    @classmethod
    def import_dict(cls, data_import: dict, ignore_checksum: bool = False) -> VAFile:
        """Imports a VAFile inclusive children from a (serialized) dictionary

        Parameters
        ----------
        data_import : dict
            [description]
        ignore_checksum: bool, optional
            If True, the checksum is ignored, defaults to False.

        Returns
        -------
        VAFile
            [description]
        """
        root = data_import.pop("__root__")
        files = {}

        for name, code in data_import.items():
            if isinstance(code, str):
                files[name] = VACode(code=code)
            else:
                files[name] = VACode(code_compressed=code, ignore_checksum=ignore_checksum)

        return VAFile(name=root, files=files)

    def write_files(
        self, path_to_target: Union[str, Path], filter: Optional[Callable[[str], str]] = None
    ):
        """Writes the VAFile and all its descendants into the given target path. The file structure is written as read from the "original"

        Parameters
        ----------
        path_to_target : Union[str, Path]
            Path to target directory
        filter : callable, optional

        """
        if not isinstance(path_to_target, Path):
            path_to_target = Path(path_to_target)

        for (name, code) in self.iter_codes():
            path_to_vafile = path_to_target / name
            path_to_vafile.parent.mkdir(exist_ok=True, parents=True)
            if filter is not None:
                text = filter(code.code)
            else:
                text = code.code
            path_to_vafile.write_text(text)

    def iter_codes(self):
        """Iterate through all name:code items"""
        return self.files.items()

    def __len__(self) -> int:
        """Number of Files inside the tree

        Returns
        -------
        int
            Length of tree
        """
        return len(self.files)

    def __eq__(self, other: object) -> bool:
        """Allows comparing two VAFiles

        Parameters
        ----------
        other : object
            Only possible for other VAFiles.

        Returns
        -------
        bool
            True if other VAFile has same root and same files.
        """
        if isinstance(other, VAFile):
            return self.root == other.root and self.files == other.files

        return NotImplemented



""" Technology class to describe all technology dependencies. Main use is scaling.

If a technology can use TRADICA, the class :class:`DMT.TRADICA.TechTradica` is recommended!

"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
from DMT.core.dut_type import DutType

try:
    from pylatex import Section
    from DMT.external.pylatex import Tex
except ImportError:
    pass


class Technology(object):
    r"""This abstract class shall bundle all relevant scaling and extraction routines that are specific for a given technologe.

    Parameters
    ----------
    name : str
        Name of the technology

    Attributes
    ----------
    name : str
        Name of the technology

    """
    name = ""

    def __init__(self, name, scaling_builder=None):
        self.name = name
        self.scaling_builder = scaling_builder

    def print_tex(self, dut_ref, mcard):
        """Prints a technology description, mainly used for autodocumentation reasons.

        Parameters
        ----------
        dut_ref : :class:`~DMT.core.DutView`
            Can be used to obtain tech quanties... (or to generate a TRADICA input file :) )
        mcard : :class:`~DMT.core.McParameterCollection`
            A Modelcard that contains all parameters that are required for scaling, as well as the parameters that shall be scaled.
        """
        doc = Tex()
        with doc.create(Section("Technology :" + self.name)):
            doc.append("Technology description missing")
        return doc

    def get_bib_entries(self):
        """bibliograpy entries of a technology"""
        return ""

    def create_mcard_library(self, lib, mcard, path, dut_type=DutType.npn):
        """Creates a file containing model cards for all sizes of duts present in the lib.

        Parameters
        ----------
        lib : :class:`~DMT.core.DutLib`
        path : str
            File path to write to.
        dut_type : :class:`~DMT.core.DutType`, optional
            Dut types to create model cards for.
        """

    def scale_all(self, mcard, lE0, bE0, nfinger, config):
        """This method receives a Modelcard (that includes relevant scaling parameters such as sheet resistances) and sets the scaled values accordingly.

        Parameters
        ----------
        mcard : :class:`~DMT.core.McParameterCollection`
            A Modelcard or McParameterCompositon that contains all parameters that are required for scaling, as well as the parameters that shall be scaled.
        lE0   : float64
            The length of the desired emitter window to be modeled by mcard.
        bE0   : float64
            The width of the desired emitter window to be modeled by mcard.
        nfinger : integer
            Number of emitter fingers.
        config : str
            A unique identifier for the configuration.
        """
        mcard = self.scale_capacitances(mcard, lE0, bE0, nfinger, config)
        mcard = self.scale_sheet_resistances(mcard, lE0, bE0, nfinger, config)
        mcard = self.scale_currents(mcard, lE0, bE0, nfinger, config)

        return mcard

    def scale_currents(self, mcard, lE0, bE0, nfinger, config):
        """This method receives a Modelcard (that includes relevant scaling parameters such as sheet resistances) and sets the scaled currents accordingly.

        Parameters
        ----------
        mcard : :class:`~DMT.core.McParameterCollection`
            A Modelcard or McParameterCompositon that contains all parameters that are required for scaling, as well as the parameters that shall be scaled.
        lE0   : float64
            The length of the desired emitter window to be modeled by mcard.
        bE0   : float64
            The width of the desired emitter window to be modeled by mcard.
        nfinger : integer
            Number of emitter fingers.
        config : str
            A unique identifier for the configuration.
        """
        raise NotImplementedError

    def scale_capacitances(self, mcard, lE0, bE0, nfinger, config):
        """This method receives a Modelcard (that includes relevant scaling parameters such as sheet resistances) and sets the scaled capacitances accordingly.

        Parameters
        ----------
        mcard : :class:`~DMT.core.McParameterCollection`
            A Modelcard or McParameterCompositon that contains all parameters that are required for scaling, as well as the parameters that shall be scaled.
        lE0   : float64
            The length of the desired emitter window to be modeled by mcard.
        bE0   : float64
            The width of the desired emitter window to be modeled by mcard.
        nfinger : integer
            Number of emitter fingers.
        config : str
            A unique identifier for the configuration.
        """
        raise NotImplementedError

    def scale_sheet_resistances(self, mcard, lE0, bE0, nfinger, config):
        """This method receives a Modelcard (that includes relevant scaling parameters such as sheet resistances) and sets the scaled resistances accordingly.

        Parameters
        ----------
        mcard : :class:`~DMT.core.McParameterCollection`
            A Modelcard or McParameterCollection that contains all parameters that are required for scaling, as well as the parameters that shall be scaled.
        lE0   : float64
            The length of the desired emitter window to be modeled by mcard.
        bE0   : float64
            The width of the desired emitter window to be modeled by mcard.
        nfinger : integer
            Number of emitter fingers.
        config : str
            A unique identifier for the configuration.
        """
        raise NotImplementedError

    def scale_modelcard(
        self, mcard, lE0, bE0, nfinger, config, lE_drawn_ref=None, bE_drawn_ref=None
    ):
        """This method scales a already finished modelcard (no sheet resistances).

        Parameters
        ----------
        mcard : :class:`~DMT.core.McParameterCollection`
            A Modelcard or McParameterCompositon that contains all parameters that are required for scaling, as well as the parameters that shall be scaled.
        lE0   : float64
            The length of the desired emitter window to be modeled by mcard.
        bE0   : float64
            The width of the desired emitter window to be modeled by mcard.
        nfinger : integer
            Number of emitter fingers.
        config : str
            A unique identifier for the configuration.
        """
        raise NotImplementedError



import requests
import numpy as np
from pathlib import Path
from DMT.core import read_data, Plot, specifiers
from DMT.core.data_reader import read_mdm


folder_path = Path(__file__).resolve().parent
folder_tmp = folder_path.parent / "tmp"


def test_read_mdm():
    df = read_data(folder_path / "test_data" / "short_freq.mdm")
    df = read_data(folder_path / "test_data" / "short_dc.mdm")


def test_read_skywater_mdm():
    req = requests.get(
        "https://raw.githubusercontent.com/google/skywater-pdk-sky130-raw-data/5738bf8753688baf2111d8d1b6a3e51a38a0bbeb/sky130_fd_pr/cells/nfet_01v8/sky130_fd_pr__nfet_01v8_w0p36u_l0p15u_m1(8701_9_10_IDVD).mdm"
    )
    if req.status_code == requests.codes.ok:
        file_tmp = folder_tmp / "skywater_nfet_01v8.mdm"
        with open(file_tmp, "w") as tmp_file:
            tmp_file.write(req.text)
        df = read_mdm(file_tmp)
        file_tmp.unlink()
    else:
        raise IOError()

    assert len(df) == 444  # 6 VG, 37 VD, 2 VB
    assert all(df["VB"].unique() == np.array([0.0, -0.9]))
    assert np.allclose(df["VG"].unique(), np.linspace(0.0, 1.8, 6))
    assert np.allclose(df["VD"].unique(), np.linspace(0.0, 1.8, 37))

    req = requests.get(
        "https://raw.githubusercontent.com/google/skywater-pdk-sky130-raw-data/5738bf8753688baf2111d8d1b6a3e51a38a0bbeb/sky130_fd_pr/cells/nfet_g5v0d10v5/sky130_fd_pr__nfet_g5v0d10v5_w0p420u_l0p500u_m1(2618_1_10_IDVD_D3).mdm"
    )
    if req.status_code == requests.codes.ok:
        file_tmp = folder_tmp / "nfet_g5v0d10v5.mdm"
        with open(file_tmp, "w") as tmp_file:
            tmp_file.write(req.text)
        df = read_mdm(file_tmp)
        file_tmp.unlink()
    else:
        raise IOError()

    assert len(df) == 1212  # 6 VG, 101 VD, 2 VB
    assert all(df["VB"].unique() == np.array([0.0, -2.5]))
    assert np.allclose(df["VG"].unique(), np.linspace(0.0, 5.0, 6))
    assert np.allclose(df["VD"].unique(), np.linspace(0.0, 5.0, 101))

    return df


if __name__ == "__main__":
    test_read_mdm()
    df = test_read_skywater_mdm()

    sp_id = specifiers.CURRENT + ["D"]
    sp_ib = specifiers.CURRENT + ["B"]
    sp_ig = specifiers.CURRENT + ["G"]
    sp_vg = specifiers.VOLTAGE + ["G"]
    sp_vd = specifiers.VOLTAGE + ["D"]
    sp_vb = specifiers.VOLTAGE + ["B"]

    plt_id = Plot(
        "ID(VD)",
        x_specifier=sp_vd,
        y_specifier=sp_id,
        legend_location="upper left",
        style="color",
        divide_by_unit=True,
    )
    plt_ib = Plot(
        "IB(VD)",
        x_specifier=sp_vd,
        y_specifier=sp_ib,
        legend_location="lower left",
        style="color",
        divide_by_unit=True,
    )
    plt_ig = Plot(
        "IG(VD)",
        x_specifier=sp_vd,
        y_specifier=sp_ig,
        legend_location="lower left",
        style="color",
        divide_by_unit=True,
    )

    for i_vb, vb, data_vb in df.iter_unique_col("VB"):
        for i_vg, vg, data_vg in data_vb.iter_unique_col("VG"):
            plt_id.add_data_set(
                data_vg["VD"],
                data_vg["ID"],
                label=f"{sp_vb.to_legend_with_value(vb)}, {sp_vg.to_legend_with_value(vg)}",
            )
            plt_ib.add_data_set(
                data_vg["VD"],
                data_vg["IB"],
                label=f"{sp_vb.to_legend_with_value(vb)}, {sp_vg.to_legend_with_value(vg)}",
            )
            plt_ig.add_data_set(
                data_vg["VD"],
                data_vg["IG"],
                label=f"{sp_vb.to_legend_with_value(vb)}, {sp_vg.to_legend_with_value(vg)}",
            )

    plt_id.plot_pyqtgraph(show=False)
    plt_ib.plot_pyqtgraph(show=False)
    plt_ig.plot_pyqtgraph(show=True)



import numpy as np
from DMT.core import DataFrame, specifiers

COL_VB = specifiers.VOLTAGE + "B"
COL_VC = specifiers.VOLTAGE + "C"
COL_VE = specifiers.VOLTAGE + "E"
COL_VBE = specifiers.VOLTAGE + ["B", "E"]
COL_VBC = specifiers.VOLTAGE + ["B", "C"]


def test_unique_iterator():
    df = DataFrame.from_dict(
        {COL_VB: np.linspace(0, 1, num=5), COL_VC: np.ones(5), COL_VE: np.zeros(5)}
    )
    df.ensure_specifier_column(COL_VBC)

    for index, vbc, data in df.iter_unique_col(COL_VBC):
        assert np.isclose(vbc, np.linspace(-1, 0, num=5)[index])


if __name__ == "__main__":
    test_unique_iterator()



""" Automatic documentation for DutLib
"""
# Copyright (C) from 2022  SemiMod
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT.
#
# DMT is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
import copy
import datetime
from pathlib import Path
from typing import Dict, List, Mapping, Sequence, Optional, Union
import numpy as np
from scipy import interpolate
from DMT.config import COMMAND_TEX, DATA_CONFIG
from DMT.core import DutType, DutLib, specifiers, sub_specifiers, specifiers_ss_para
from DMT.core.plot import MIX, PLOT_STYLES, natural_scales, Plot
from DMT.external.os import recursive_copy, rmtree

try:
    from pylatex import Section, Subsection, SmallText, Tabular, NoEscape, Center, Figure
    from pylatex.base_classes import Arguments
    from DMT.external.pylatex import SubFile, Tex, CommandInput, CommandLabel
except ImportError:
    pass

# defaults for autodoc feature plots
PLOT_DEFAULTS = {
    DutType.npn: {
        "gummel_vbc": {
            "x_log": False,
            "y_log": True,
            "at": specifiers.VOLTAGE + "B" + "C" + sub_specifiers.FORCED,
            "quantity_x": specifiers.VOLTAGE + "B" + "E",
            "quantity_y": specifiers.CURRENT_DENSITY + "C",
            "legend_location": "upper left",
            "y_limits": (None, None),
            "x_limits": (None, None),
            "tex": r"Gummel @ $V_{\mathrm{BC}}$.",
        },
        "gummel_vbc_mark_ft": {
            "x_log": False,
            "y_log": True,
            "at": specifiers.VOLTAGE + "B" + "C" + sub_specifiers.FORCED,
            "quantity_x": specifiers.VOLTAGE + "B" + "E",
            "quantity_y": specifiers.CURRENT_DENSITY + "C",
            "legend_location": "upper left",
            "y_limits": (None, None),
            "x_limits": (None, None),
            "tex": r"Gummel @ $V_{\mathrm{BC}}$.",
        },
        "output_vbe": {
            "x_log": False,
            "y_log": False,
            "at": specifiers.VOLTAGE + "B" + "E" + sub_specifiers.FORCED,
            "quantity_x": specifiers.VOLTAGE + "C" + "E",
            "quantity_y": specifiers.CURRENT_DENSITY + "C",
            "legend_location": "upper left",
            "y_limits": (None, None),
            "x_limits": (None, None),
            "tex": r"Output @ $V_{\mathrm{BE}}$.",
        },
        "output_ib": {
            "x_log": False,
            "y_log": False,
            "at": specifiers.CURRENT_DENSITY + "B",
            "quantity_x": specifiers.VOLTAGE + "C" + "E",
            "quantity_y": specifiers.CURRENT_DENSITY + "C",
            "legend_location": "upper left",
            "y_limits": (None, None),
            "x_limits": (None, None),
            "tex": r"Output @ $I_{\mathrm{B}}$.",
        },
        "ft_jc_vbc": {
            "x_log": True,
            "y_log": False,
            "at": specifiers.VOLTAGE + "B" + "C" + sub_specifiers.FORCED,
            "quantity_x": specifiers.CURRENT_DENSITY + "C",
            "quantity_y": specifiers.TRANSIT_FREQUENCY,
            "legend_location": "upper left",
            "y_limits": (0, None),
            "x_limits": (None, None),
            "tex": r"$f_{\mathrm{T}} \left( J_{\mathrm{C}} \right) $ @ $V_{\mathrm{BC}}$.",
        },
        "fmax_jc_vbc": {
            "x_log": True,
            "y_log": False,
            "at": specifiers.VOLTAGE + "B" + "C" + sub_specifiers.FORCED,
            "quantity_x": specifiers.CURRENT_DENSITY + "C",
            "quantity_y": specifiers.MAXIMUM_OSCILLATION_FREQUENCY,
            "legend_location": "upper left",
            "y_limits": (0, None),
            "x_limits": (None, None),
            "tex": r"$f_{\mathrm{max}} \left( J_{\mathrm{C}} \right) $ @ $V_{\mathrm{BC}}$.",
        },
        "ft_jc_vce": {
            "x_log": True,
            "y_log": False,
            "at": specifiers.VOLTAGE + "C" + "E" + sub_specifiers.FORCED,
            "quantity_x": specifiers.CURRENT_DENSITY + "C",
            "quantity_y": specifiers.TRANSIT_FREQUENCY,
            "legend_location": "upper left",
            "y_limits": (0, None),
            "x_limits": (None, None),
            "tex": r"$F_{\mathrm{T}} \left( J_{\mathrm{C}} \right) $ @ $V_{\mathrm{CE}}$.",
        },
        "fmax_jc_vce": {
            "x_log": True,
            "y_log": False,
            "at": specifiers.VOLTAGE + "C" + "E" + sub_specifiers.FORCED,
            "quantity_x": specifiers.CURRENT_DENSITY + "C",
            "quantity_y": specifiers.MAXIMUM_OSCILLATION_FREQUENCY,
            "legend_location": "upper left",
            "y_limits": (0, None),
            "x_limits": (None, None),
            "tex": r"$f_{\mathrm{max}} \left( J_{\mathrm{C}} \right) $ @ $V_{\mathrm{CE}}$.",
        },
        "beta_jc_vbc": {
            "x_log": True,
            "y_log": False,
            "at": specifiers.VOLTAGE + "B" + "C" + sub_specifiers.FORCED,
            "quantity_x": specifiers.CURRENT_DENSITY + "C",
            "quantity_y": specifiers.DC_CURRENT_AMPLIFICATION,
            "legend_location": "upper left",
            "y_limits": (None, None),
            "x_limits": (None, None),
            "tex": r"$\beta_{\mathrm{DC}} \left( J_{\mathrm{C}} \right)$ @ $V_{\mathrm{BC}}$.",
        },
        "rey21_f_vbe_vbc": {
            "x_log": True,
            "y_log": False,
            "at": [
                specifiers.VOLTAGE + "B" + "E" + sub_specifiers.FORCED,
                specifiers.VOLTAGE + "B" + "C" + sub_specifiers.FORCED,
            ],
            "quantity_x": specifiers.FREQUENCY,
            "quantity_y": specifiers.SS_PARA_Y + "C" + "B" + sub_specifiers.REAL,
            "legend_location": "upper left",
            "y_limits": (None, None),
            "x_limits": (None, None),
            "tex": r"$\Re \left\{ Y_{21} \right\} \left( f \right)$ @ $V_{\mathrm{BC}} @ V_{\mathrm{BE}}$.",
        },
        "imy11_f_vbe_vbc": {
            "x_log": True,
            "y_log": True,
            "at": [
                specifiers.VOLTAGE + "B" + "E" + sub_specifiers.FORCED,
                specifiers.VOLTAGE + "B" + "C" + sub_specifiers.FORCED,
            ],
            "quantity_x": specifiers.FREQUENCY,
            "quantity_y": specifiers.SS_PARA_Y + "B" + "B" + sub_specifiers.IMAG,
            "legend_location": "upper left",
            "y_limits": (None, None),
            "x_limits": (None, None),
            "tex": r"$\Im \left\{ Y_{11} \right\} \left( f \right)$ @ $V_{\mathrm{BC}} @ V_{\mathrm{BE}}$.",
        },
        "y21_jc_vbc": {
            "x_log": True,
            "y_log": True,
            "at": [
                specifiers.FREQUENCY,
                specifiers.VOLTAGE + "B" + "C" + sub_specifiers.FORCED,
            ],
            "quantity_x": specifiers.CURRENT_DENSITY + "C",
            "quantity_y": specifiers.SS_PARA_Y + "C" + "B" + sub_specifiers.REAL,
            "legend_location": "upper left",
            "y_limits": (None, None),
            "x_limits": (None, None),
            "tex": r"$\Re \left\{ Y_{21} \right\} \left( J_{\mathrm{C}} \right)$ @ $V_{\mathrm{BC}}$.",
        },
        "y21_jc_vbc_mark_ft": {
            "x_log": True,
            "y_log": True,
            "at": [
                specifiers.FREQUENCY,
                specifiers.VOLTAGE + "B" + "C" + sub_specifiers.FORCED,
            ],
            "quantity_x": specifiers.CURRENT_DENSITY + "C",
            "quantity_y": specifiers.SS_PARA_Y + "C" + "B" + sub_specifiers.REAL,
            "legend_location": "upper left",
            "y_limits": (None, None),
            "x_limits": (None, None),
            "tex": r"$\Re \left\{ Y_{21} \right\} \left( J_{\mathrm{C}} \right)$ @ $V_{\mathrm{BC}}$.",
        },
        "y21_jc_vbc_mark_ft": {
            "x_log": True,
            "y_log": True,
            "at": [
                specifiers.FREQUENCY,
                specifiers.VOLTAGE + "B" + "C" + sub_specifiers.FORCED,
            ],
            "quantity_x": specifiers.CURRENT_DENSITY + "C",
            "quantity_y": specifiers.SS_PARA_Y + "C" + "B" + sub_specifiers.REAL,
            "legend_location": "upper left",
            "y_limits": (None, None),
            "x_limits": (None, None),
            "tex": r"$\Re \left\{ Y_{21} \right\} \left( J_{\mathrm{C}} \right)$ @ $V_{\mathrm{BC}}$.",
        },
        "tj_jc_at_vbc": {
            "x_log": True,
            "y_log": False,
            "at": specifiers.VOLTAGE + "B" + "C" + sub_specifiers.FORCED,
            "quantity_x": specifiers.CURRENT_DENSITY + "C",
            "quantity_y": specifiers.TEMPERATURE,
            "legend_location": "upper left",
            "y_limits": (None, None),
            "x_limits": (None, None),
            "tex": r"$T_{\mathrm{j}} \left( J_{\mathrm{C}} \right)$ @ $V_{\mathrm{BC}}$.",
            "rth": 3e3,
        },
    },
    DutType.n_mos: {
        "id(vg)": {
            "x_log": False,
            "y_log": False,
            "at": [specifiers.VOLTAGE + ["D"], specifiers.VOLTAGE + ["B"]],
            "quantity_x": specifiers.VOLTAGE + ["G"],
            "quantity_y": specifiers.CURRENT + ["D"],
            "legend_location": "upper left",
            "y_limits": (None, None),
            "x_limits": (None, None),
            "tex": r"$I_{\mathrm{D}}(V_{\mathrm{G}})@V_{\mathrm{D}}$.",
        },
        "id(vd)": {
            "x_log": False,
            "y_log": False,
            "at": [specifiers.VOLTAGE + ["G"], specifiers.VOLTAGE + ["B"]],
            "quantity_x": specifiers.VOLTAGE + ["D"],
            "quantity_y": specifiers.CURRENT + ["D"],
            "legend_location": "upper left",
            "y_limits": (None, None),
            "x_limits": (None, None),
            "tex": r"$I_{\mathrm{D}}(V_{\mathrm{D}})@V_{\mathrm{G}}$.",
        },
    },
    DutType.p_mos: {
        "id(vg)": {
            "x_log": False,
            "y_log": False,
            "at": [specifiers.VOLTAGE + ["D"], specifiers.VOLTAGE + ["B"]],
            "quantity_x": specifiers.VOLTAGE + ["G"],
            "quantity_y": specifiers.CURRENT + ["D"],
            "legend_location": "upper left",
            "y_limits": (None, None),
            "x_limits": (None, None),
            "tex": r"$I_{\mathrm{D}}(V_{\mathrm{G}})@V_{\mathrm{D}}$.",
        },
        "id(vd)": {
            "x_log": False,
            "y_log": False,
            "at": [specifiers.VOLTAGE + ["G"], specifiers.VOLTAGE + ["B"]],
            "quantity_x": specifiers.VOLTAGE + ["D"],
            "quantity_y": specifiers.CURRENT + ["D"],
            "legend_location": "upper left",
            "y_limits": (None, None),
            "x_limits": (None, None),
            "tex": r"$I_{\mathrm{D}}(V_{\mathrm{D}})@V_{\mathrm{G}}$.",
        },
    },
}


class DocuDutLib(object):
    """Documentation of an DutLib

    Parameters
    ----------
    dut_lib : :class:`~DMT.core.DutLib`
        Library to document
    devices : Sequence[Mapping[str, object]]
        This list specifies which devices to select from the lib. Only needed if "dev_mode"="sel" in the mode argument.
        Each dict specifies a contact configuration, length and width. All devices that match these properties are used for plotting.
        Example::

            devices = [{
                'contact_config':'CBEBC',
                'length'        :2.8e-6,
                'width'         :0.22e-6,
            },]
    date : convertable to str
        Date of the documentation report.
    """

    def __init__(
        self, dut_lib: DutLib, devices: Optional[Sequence[Mapping[str, object]]] = None, date=None
    ):
        self.dut_lib = dut_lib

        if devices is None:
            self.duts = self.dut_lib.duts
        else:
            self.duts = []  # array of DMT Duts
            for device_specs in devices:
                for dut in self.dut_lib:
                    # check all properties of device_spec
                    ok = True
                    for device_spec, val in device_specs.items():
                        if isinstance(val, str):
                            dut_property = getattr(dut, device_spec)
                            if not val in dut_property:
                                ok = False
                        elif isinstance(val, float):
                            dut_property = getattr(dut, device_spec)
                            try:
                                if not np.isclose(dut_property, val):
                                    ok = False
                            except ValueError:  # tetrodes will not work like this (tuple dut_property)
                                pass

                    if ok:
                        print("Found device " + dut.name + ".")
                        self.duts.append(dut)

        if not self.duts:
            raise IOError("Found 0 devices.")

        print("Search finished...\nFound " + str(len(self.duts)) + " devices for plotting.")

        if date is None:
            self.date = datetime.datetime.today().strftime("%Y-%m-%d")
        else:
            self.date = date

        self.plts = []

    def generate_docu(
        self,
        target_base_path: Union[str, Path],
        plot_specs: List[Dict],
        save_tikz_settings: Optional[Dict] = None,
        show: bool = False,
    ):
        """Generate the full documentation folder with all plots and files.

        Parameters
        ----------
        target_base_path : str | Path
            Target path
        plot_specs : List[Dict]
            Plot specs passed to :meth:`~DMT.core.docu_dut_lib.DocuDutLib.create_all_plots`
        save_tikz_settings : Dict
            Plot specs passed to :meth:`~DMT.core.docu_dut_lib.DocuDutLib.generate_all_plots`
        show : bool, optional
            If True, the plots are displayed before the documentation folder is created, by default False
        """
        self.create_all_plots(plot_specs)

        # TODO
        # technology
        # mcard
        # ADD data from mcard simulation

        if show:
            if len(self.plts) > 1:
                for plt in self.plts[:-1]:
                    plt.plot_pyqtgraph(show=False)

            self.plts[-1].plot_pyqtgraph(show=True)

        self.generate_all_plots(target_base_path, save_tikz_settings)

        self.create_documentation(target_base_path)

    def create_all_plots(
        self,
        plot_specs: List[Dict],
    ):
        """This method generates plots for selected devices in the DutLib in a highly configurable way.

        Parameters
        ----------
        plot_specs : [dict]
            For every plot type you want to create, this list contains one dict. Example::

                {
                    'type'        : 'ft_jc_vbc', #possible values are the plot_types that are stored in the defaults (see code below)
                    'exclude_at'  : [-0.2,0.5],  #every plot type generates lines "at" some quantity, e.g. gummel_vbc generates for every VBC. With this argument you can exclude lines.
                    'key'         : 'freq_vbc',  #data keys that will be used for this plot
                    'exact_match' : False,       #Bool, if True, use only keys that match "key" exactly
                    'FREQ'        : 10e9,        #optional, float: If given, use only data at FREQ
                    'xmin'        : 10e-2,       #optional, float: Minimum Value on x-axis to be displayed
                    'xmax'        : 1e2,         #optional, float: Maximum Value on x-axis to be displayed
                    'ymax'        : 300,         #optional, float: Maximum Value on y-axis to be displayed
                    'ymin'        : 0,           #optional, float: Minimum Value on y-axis to be displayed
                    'no_at'       : True,        #optional, Bool: if True, do not display the "at" quantities in the legend.
                },

        """
        plot_spec_defaults = {
            "style": MIX,
            "legend": True,
            "dut_type": DutType.npn,
            "no_at": False,  # no at_specifier= in legend
        }

        # set defaults
        for plot_spec in plot_specs:
            for key in plot_spec_defaults:
                if key not in plot_spec.keys():
                    plot_spec[key] = plot_spec_defaults[key]

        # check plot_spec
        for plot_spec in plot_specs:
            try:
                plot_type = plot_spec["type"]
            except KeyError as err:
                raise IOError("type not specified for plot specification.") from err

            valid_plots = PLOT_DEFAULTS[plot_spec["dut_type"]].keys()
            if not plot_type in valid_plots:
                raise IOError(
                    "Plot type "
                    + plot_type
                    + " not valid. Valid: "
                    + " ".join(str(PLOT_DEFAULTS.keys()))
                    + " ."
                )

            if not plot_spec["style"] in PLOT_STYLES:
                raise IOError("Plot style not valid. Valid: " + " ".join(PLOT_STYLES))

            if not "key" in plot_spec.keys():
                raise IOError("Database key not specified for plot of type " + plot_type + " .")

            # matching key?
            if "exact_match" not in plot_spec.keys():
                plot_spec["exact_match"] = False

        # ensure that plot with type gummel_vbc_mark_ft comes first
        for i, plot_spec in enumerate(plot_specs):
            if plot_spec["type"] == "gummel_vbc_mark_ft":
                if i == 0:
                    pass
                else:
                    plot_specs[0], plot_specs[i] = plot_specs[i], plot_specs[0]  # swap elements

                break

        print("Generating plots...")
        # for every plot try to generate appropriate plots
        self.plts = []
        for plot_spec in plot_specs:
            plot_type = plot_spec["type"]
            print("Generating plots of type " + plot_type + " .")

            style = plot_spec["style"]
            print("Chosen plot style: " + style)

            for dut in self.duts:
                # load default settings of plot
                try:
                    x_log = PLOT_DEFAULTS[dut.dut_type][plot_type]["x_log"]
                    y_log = PLOT_DEFAULTS[dut.dut_type][plot_type]["y_log"]
                    quantity_x = PLOT_DEFAULTS[dut.dut_type][plot_type]["quantity_x"]
                    quantity_y = PLOT_DEFAULTS[dut.dut_type][plot_type]["quantity_y"]
                    legend_location = PLOT_DEFAULTS[dut.dut_type][plot_type]["legend_location"]
                    at_specifier = PLOT_DEFAULTS[dut.dut_type][plot_type]["at"]
                except KeyError:
                    continue  # no plot_type in plot_defaults for this plot_spec

                # overwrite defaults with plot_spec
                try:
                    legend_location = plot_spec["legend_location"]
                except KeyError:
                    pass

                if not isinstance(at_specifier, list):
                    at_specifier = [at_specifier]

                quantities_to_ensure = [quantity_x, quantity_y] + at_specifier
                if "mark_ft" in plot_type:
                    quantities_to_ensure.append(specifiers.TRANSIT_FREQUENCY)
                    peaks = {"vbe": [], "jc": [], "vbc": []}  # store peak ft values for later

                at_scale = []
                for at_ in at_specifier:
                    try:
                        at_scale_ = natural_scales[at_.specifier]
                    except AttributeError:
                        at_scale_ = natural_scales[at_]

                    if at_ == specifiers.CURRENT + "B" or at_ == specifiers.CURRENT_DENSITY + "B":
                        at_scale_ = at_scale_ * 1e3

                    at_scale.append(at_scale_)

                print("Generating plot of type " + plot_type + " for dut " + dut.name + " ...")
                name = [
                    "dut_",
                    dut.name,
                    "_",
                    plot_type,
                ]
                if specifiers.TEMPERATURE in plot_spec.keys():
                    name.append("atT" + str(plot_spec[specifiers.TEMPERATURE]) + "K")
                if specifiers.FREQUENCY in plot_spec.keys():
                    name.append("atf" + str(plot_spec[specifiers.FREQUENCY] * 1e-9) + "GHz")

                for at_ in at_specifier:
                    name.append("at" + at_)

                name = "_".join(name)

                # calc drawn emitter windows area
                AE0_drawn = dut.width * dut.length * dut.contact_config.count("E")

                # find temperatures
                temps = []
                for key in dut.data.keys():
                    temps.append(dut.get_key_temperature(key))
                temps = list(set(temps))

                for temp in temps:

                    y_scale = 1
                    x_label = None  # autolabel
                    y_label = None
                    if (
                        quantity_y.specifier in specifiers.SS_PARA_Y
                    ):  # special cases that I do not want in DMT
                        y_scale = 1e3 / (1e6 * 1e6)  # mS/um^2
                        y_label = r"$\Re{ \left\{ Y_{21} \right\} } / \si{\milli\siemens\per\square\micro\meter } $"

                    plt = Plot(
                        name,
                        style=style,
                        num=name,
                        x_specifier=quantity_x,
                        y_specifier=quantity_y,
                        x_log=x_log,
                        y_log=y_log,
                        y_scale=y_scale,
                        x_label=x_label,
                        y_label=y_label,
                        legend_location=legend_location,
                    )
                    plt.dut_name = (
                        dut.name
                        + f"w{dut.width*1e6:.2f}um_l{dut.length*1e6:.2f}um_{len(self.plts)}"
                    )
                    plt.plot_type = plot_type
                    plt.dut = dut
                    plt.temp = temp
                    plt.plot_spec = plot_spec

                    n = 0
                    for key in dut.data.keys():
                        # selected only keys at temp
                        if not dut.get_key_temperature(key) == temp:
                            continue

                        if specifiers.TEMPERATURE in plot_spec.keys():
                            if temp != plot_spec[specifiers.TEMPERATURE]:
                                continue  # key not suitable

                        match = False
                        if plot_spec["exact_match"]:
                            match = plot_spec["key"] == dut.split_key(key)[-1]
                        else:
                            match = plot_spec["key"] in key

                        if match:
                            df = dut.data[key]
                            if specifiers.FREQUENCY in plot_spec.keys():
                                try:
                                    df = df[
                                        df[specifiers.FREQUENCY] == plot_spec[specifiers.FREQUENCY]
                                    ]
                                except KeyError:
                                    pass

                            for quantity in quantities_to_ensure:
                                try:
                                    df.ensure_specifier_column(
                                        quantity, area=AE0_drawn, ports=dut.ac_ports
                                    )
                                except KeyError:
                                    if quantity == specifiers.TEMPERATURE:
                                        # calculate rough temperature
                                        pdiss = (
                                            df[specifiers.CURRENT + "C"].to_numpy()
                                            * df[specifiers.VOLTAGE + "C"].to_numpy()
                                        )

                                        # for the time beeing, only works for CBEBC devices...
                                        # calculate rth parameter
                                        a = (
                                            4.0
                                            * self.dut_lib.dut_ref.length
                                            / self.dut_lib.dut_ref.width
                                        )
                                        F_th = 1
                                        if a > 0.0:
                                            F_th = self.dut_lib.dut_ref.length / np.log(a)
                                        SRTHRM = plot_spec["rth"] / F_th

                                        # scale
                                        a = 4.0 * dut.length / dut.width
                                        F_th = 1
                                        if a > 0.0:
                                            F_th = dut.length / np.log(a)
                                        rth = SRTHRM * F_th

                                        df.loc[:, quantity] = temp + pdiss * rth
                                        dut.rth = rth

                                try:
                                    if quantity.specifier in specifiers_ss_para.SS_PARA_Y:
                                        df.loc[:, quantity] = df[quantity] / AE0_drawn
                                except:
                                    pass

                            at_vals = []
                            for i, at_ in enumerate(at_specifier):
                                at_val = df[at_].to_numpy()
                                if at_.specifier == specifiers.VOLTAGE:
                                    at_val = np.round(at_val, decimals=3)
                                    at_val = np.unique(at_val)
                                elif at_.specifier == specifiers.CURRENT:
                                    at_val = np.round(at_val, decimals=8)
                                    at_val = np.unique(at_val)
                                elif at_.specifier == specifiers.CURRENT_DENSITY:
                                    at_val = np.round(at_val, decimals=0)
                                    at_val = np.unique(at_val)

                                if "at_vals" in plot_spec:
                                    at_val = plot_spec["at_vals"][i]
                                    if not isinstance(at_val, list):
                                        at_val = [at_val]

                                at_vals.append(at_val)

                            units = []
                            for i, at_ in enumerate(at_specifier):
                                units.append(at_.get_tex_unit(scale=at_scale[i]))

                            f = []
                            if len(at_specifier) == 1:
                                for point in at_vals[0]:
                                    f.append((point,))
                            elif len(at_specifier) == 2:
                                f = [(x, y) for x in at_vals[0] for y in at_vals[1]]
                            else:
                                raise IOError("at with more than two specifiers not implemented.")

                            for point in f:
                                df_filter = True
                                at_str = r"$"
                                for i, (speci, u, scale_) in enumerate(
                                    zip(at_specifier, units, at_scale)
                                ):
                                    df_filter = np.logical_and(
                                        df_filter, np.isclose(df[speci], point[i], rtol=1e-3)
                                    )
                                    if at_str != r"$":
                                        at_str += r",\,"
                                    if plot_spec["no_at"]:
                                        at_str += r"{0:1.2f}".format(point[i] * scale_) + u
                                    else:
                                        at_str += (
                                            speci.to_tex()
                                            + r" = {0:1.2f}".format(point[i] * scale_)
                                            + u
                                        )

                                at_str += r"$"
                                df_tmp = df[df_filter]
                                x = df_tmp[quantity_x].to_numpy()
                                y = df_tmp[quantity_y].to_numpy()

                                # device if legend is wanted...default yes
                                kwargs = {}
                                if plot_spec["legend"]:
                                    kwargs["label"] = at_str

                                plt.add_data_set(
                                    x,
                                    y,
                                    **kwargs,
                                )

                                # add dots at peak ft
                                if "mark_ft" in plot_type:
                                    ft = df_tmp[specifiers.TRANSIT_FREQUENCY].to_numpy() * 1e-9
                                    interp_fun_ft = interpolate.interp1d(x, ft)
                                    interp_fun_ic = interpolate.interp1d(x, y)

                                    index_peak_ft = np.argmax(ft)
                                    try:
                                        vbe_new = np.linspace(
                                            x[index_peak_ft - 10], x[index_peak_ft + 10], 201
                                        )  # may error
                                    except IndexError:
                                        vbe_new = np.linspace(
                                            x[index_peak_ft - 10], x[-1], 201
                                        )  # may error

                                    index_peak_ft = np.argmax(interp_fun_ft(vbe_new))

                                    vbe_peak = vbe_new[index_peak_ft]
                                    jc_peak = interp_fun_ic(vbe_new[index_peak_ft])
                                    vbc_peak = point

                                    plt.add_data_set(
                                        np.tile(np.array(vbe_peak), 5),
                                        np.tile(np.array(jc_peak), 5),
                                        style=" ök",  # black dots
                                    )
                                    peaks["vbe"].append(np.tile(vbe_peak, 10))
                                    peaks["jc"].append(np.tile(jc_peak, 10))
                                    peaks["vbc"].append(np.tile(point[0], 10))

                                n = n + 1

                            if "ymin" in plot_spec.keys() or "ymax" in plot_spec.keys():
                                if not "ymin" in plot_spec.keys():
                                    plot_spec["ymin"] = None
                                if not "ymax" in plot_spec.keys():
                                    plot_spec["ymax"] = None
                                plt.y_limits = (plot_spec["ymin"], plot_spec["ymax"])

                            else:
                                try:
                                    plt.y_limits = PLOT_DEFAULTS[plot_type]["y_limits"]
                                except KeyError:
                                    pass

                            if "xmin" in plot_spec.keys() or "xmax" in plot_spec.keys():
                                if not "xmin" in plot_spec.keys():
                                    plot_spec["xmin"] = None
                                if not "xmax" in plot_spec.keys():
                                    plot_spec["xmax"] = None
                                plt.x_limits = (plot_spec["xmin"], plot_spec["xmax"])

                            else:
                                try:
                                    plt.x_limits = PLOT_DEFAULTS[plot_type]["x_limits"]
                                except KeyError:
                                    pass

                    # plots that required to mark peak of ft are accounted for here.
                    if plot_type == "gummel_vbc_mark_ft":
                        dut.peaks = peaks
                    elif plot_type == "output_ib":  # add ft peaks if they exist
                        try:
                            peaks = dut.peaks
                            jc = np.array(peaks["jc"]).flatten()
                            vbe = np.array(peaks["vbe"]).flatten()
                            vbc = np.array(peaks["vbc"]).flatten()
                            vce = vbe - vbc

                            inds = vce.argsort()
                            plt.add_data_set(
                                vce[inds],
                                jc[inds],
                                style="-ök",  # black dots
                            )
                        except AttributeError:
                            pass

                    print(
                        "...finished plot of type "
                        + plot_type
                        + " for dut "
                        + dut.name
                        + " , found "
                        + str(n)
                        + " lines."
                    )

                    if n == 0:
                        print("Found no lines for plot " + plot_type + " .")
                    else:
                        self.plts.append(plt)

    def generate_all_plots(
        self, target_base_path: Union[str, Path], save_tikz_settings: Optional[Dict] = None
    ):
        """Generates the plot tex and pdf files.

        If a plot with the same name already exists, it is deleted before the new is created.

        Parameters
        ----------
        target_base_path : str | Path
            Target base path. Inside this folder the plots will be placed in a "figs" subfolder.
        save_tikz_settings : Dict
            This dict specifies how the plots will be stored, they are directly used as parameters for :meth:`~DMT.core.Plot.save_tikz`. Example::

                save_tikz_settings = {
                    'width'      : '4.5in', #width of the Tikz Pictures
                    'height'     : '4.5in', #height of the Tikz Pictures
                    'fontsize'   : 'Large', #Fontsize Tex specification
                    'clean'      : True,    #If True: remove all files except rendered picture after build
                    'svg'        : False    #bool, False: If True, build svg files instead of pdf files.
                    'build'      : True     #bool, True: If True, build the Tex files using pdflatex compiler. Else only print .tex files.
                    'mark_repeat': 20       #int,20: Only show every nth marker, where n=mark_repeat.
                    'clean'      : False,   #bool, False: Remove all files except *.pdf files in plots. Schroeter likes this.
                }

        """
        if isinstance(target_base_path, Path):
            base_path = target_base_path
        else:
            base_path = Path(target_base_path)

        save_tikz_settings_defaults = {
            "width": "3in",
            "height": "5in",
            "standalone": True,
            "svg": False,
            "build": True,
            "mark_repeat": 20,
            "clean": False,  # Remove all files except *.pdf files in plots
        }

        if save_tikz_settings is None:
            save_tikz_settings = save_tikz_settings_defaults
        else:
            for key in save_tikz_settings_defaults:
                if not key in save_tikz_settings.keys():
                    save_tikz_settings[key] = save_tikz_settings[key]

        for plt in self.plts:
            plot_path = (
                base_path / "figs" / plt.dut_name / ("T" + str(plt.temp) + "K") / plt.plot_type
            )

            if plt.plot_type == "tj_jc_at_vbc":
                plot_path = (
                    base_path
                    / "figs"
                    / plt.dut_name
                    / ("T" + str(plt.temp) + "K")
                    / (plt.plot_type + "rth_" + "{0:1.1f}".format(plt.dut.rth * 1e-3) + "kWperK")
                )

            if plot_path.exists():
                rmtree(plot_path)

            plot_path.mkdir(parents=True, exist_ok=True)

            # special output from plot_spec
            save_tikz_settings_tmp = copy.deepcopy(save_tikz_settings)
            for special in ["width", "height"]:
                try:
                    save_tikz_settings_tmp[special] = plt.plot_spec[special]
                except KeyError:
                    continue

            plt.legend_location = "upper right outer"
            filename = plt.save_tikz(plot_path, **save_tikz_settings_tmp)

            plt.path = plot_path / filename.replace("tex", "pdf")

    def create_documentation(self, target_base_path: Union[str, Path]):
        """Generates the other tex files from the template.

        The template path from the config is used. The key is::

            directories:
                libautodoc: null

        If None/null, the DMT supplied template is used.

        Parameters
        ----------
        target_base_path : str | Path
            Target path for the documentation report.
        """
        if isinstance(target_base_path, Path):
            destination = target_base_path
        else:
            destination = Path(target_base_path)

        dir_source = DATA_CONFIG["directories"]["libautodoc"]

        destination.mkdir(parents=True, exist_ok=True)

        recursive_copy(dir_source, destination)

        try:
            # now rename _x_title and _author
            string_deckblatt = (destination / "content" / "deckblatt.tex").read_text()
            string_deckblatt = string_deckblatt.replace("_author", DATA_CONFIG["user_name"])
            string_deckblatt = string_deckblatt.replace(
                "_company", str(DATA_CONFIG["company"]).replace("_", r"\_")
            )
            string_deckblatt = string_deckblatt.replace(
                "_x_title", str(DATA_CONFIG["docu_topic"]).replace("_", r"\_")
            )
            string_deckblatt = string_deckblatt.replace(
                "_wafer", str(self.dut_lib.wafer).replace("_", r"\_")
            )
            string_deckblatt = string_deckblatt.replace(
                "_date_TO", str(self.dut_lib.date_tapeout).replace("_", r"\_")
            )
            string_deckblatt = string_deckblatt.replace(
                "_date_received", str(self.dut_lib.date_received).replace("_", r"\_")
            )
            string_deckblatt = string_deckblatt.replace(
                "_date_docu", str(self.date).replace("_", r"\_")
            )

            (destination / "content" / "deckblatt.tex").write_text(string_deckblatt)
        except FileNotFoundError:
            pass

        # finish copy template

        # subfile that contains overview of library
        lib_tex = SubFile(master="../documentation.tex")
        lib_tex.append(self.dut_lib.toTex())

        # subfile that contains all plots
        plots_tex = SubFile(master="../documentation.tex")
        doc = Tex()
        for dut in self.duts:
            with doc.create(Section(dut.name)):
                plts_for_this_dut = []
                for plt in self.plts:
                    if plt.dut == dut:
                        plts_for_this_dut.append(plt)

                temps = []
                for plt in plts_for_this_dut:
                    temps.append(plt.temp)

                temps = list(set(temps))

                for temp in temps:
                    with doc.create(Subsection("T=" + str(temp) + "K")):
                        for plt in plts_for_this_dut:
                            if plt.temp == temp:
                                # check if plot exists:
                                # Why? The plots are added as images not as tikz plots Oo
                                if plt.path.is_file():
                                    doc.append(NoEscape(r"\FloatBarrier "))
                                    with doc.create(Figure(position="ht!")) as _plot:
                                        _plot.append(
                                            NoEscape(r"\setlength\figurewidth{\textwidth}")
                                        )
                                        # _plot.append(CommandInput(arguments=Arguments(plt.path)))
                                        _plot.add_image('"' + str(plt.path) + '"')
                                        _plot.add_caption(
                                            NoEscape(
                                                PLOT_DEFAULTS[dut.dut_type][plt.plot_type]["tex"]
                                            )
                                        )
                                        # _plot.append(CommandLabel(arguments=Argument(plt.dut_name + plt.)))

                                    doc.append(NoEscape(r"\FloatBarrier "))

        # put into subfile
        plots_tex.append(doc)

        lib_tex.generate_tex(str(destination / "content" / "lib_overview"))
        plots_tex.generate_tex(str(destination / "content" / "lib_plots"))
